/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*                                                                 */
/*                OPERATIONS DANS LES CORPS DE NOMBRES             */
/*                                                                 */
/*                                                                 */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

#include "genpari.h"

/*
  Operations sur les elements de corps de nombres.
  Un element sera represente par un vecteur colonne dans la base d'entiers
  nf[7].
  */


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*                                                                 */
/*                  OPERATIONS SUR LES ELEMENTS                    */
/*                                                                 */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

int
isnfscalar(GEN x)
{
  long lx=lg(x),i;
  
  for(i=2;(i<lx)&&gcmp0((GEN)x[i]);i++);
  return (i==lx);
}

GEN
element_mul(GEN nf, GEN x, GEN y)
                
/* Recoit deux vecteurs de longueur N representant x et y dans la base 
 d'entiers (peut etre modulo p) et ressort leur produit sous forme d'un
 vecteur. */

{
  long av=avma,tetpil,i,j,k,N,tx=typ(x),ty=typ(y);
  GEN s,v,c,p1;

  nf=checknf(nf);N=lgef((GEN)nf[1])-3;
  if(tx==9)
  {
    if(gegal((GEN)x[1],(GEN)nf[1])) x=(GEN)x[2];
    else err(talker,"not the same polynomial in a number field operation");
  }
  if(ty==9)
  {
    if(gegal((GEN)y[1],(GEN)nf[1])) y=(GEN)y[2];
    else err(talker,"not the same polynomial in a number field operation");
  }
  if(tx<=10)
  {
    if(ty<=10) p1=gmul(x,y);
    else
    {
      if(ty!=18) err(talker,"incorrect type in nfmul");
      p1=gmul(x,gmul((GEN)nf[7],y));
    }
    tetpil=avma;return gerepile(av,tetpil,algtobasis(nf,p1));
  }
  else
  {
    if(ty<=10)
    {
      if(tx!=18) err(talker,"incorrect type in nfmul");
      p1=gmul(gmul((GEN)nf[7],x),y);
      tetpil=avma;return gerepile(av,tetpil,algtobasis(nf,p1));
    }
    else
    {
      if(isnfscalar(x)) return gmul((GEN)x[1],y);
      if(isnfscalar(y)) return gmul((GEN)y[1],x);
      v=cgetg(N+1,18);
      for(k=1;k<=N;k++)
      {
	s=gzero;
	for(i=1;i<=N;i++)
	{
	  c=gcoeff((GEN)nf[9],k,(i-1)*N+i);
	  if(signe(c))
	  {
	    if(gcmp1(c)) s=gadd(s,gmul((GEN)x[i],(GEN)y[i]));
	    else s=gadd(s,gmul(gmul((GEN)x[i],(GEN)y[i]),c));
	  }
	  for(j=i+1;j<=N;j++)
	  {
	    c=gcoeff((GEN)nf[9],k,(i-1)*N+j);
	    if(signe(c))
	    {
	      if(gcmp1(c)) s=gadd(s,gadd(gmul((GEN)x[i],(GEN)y[j]),gmul((GEN)x[j],(GEN)y[i])));
	      else s=gadd(s,gmul(gadd(gmul((GEN)x[i],(GEN)y[j]),gmul((GEN)x[j],(GEN)y[i])),c));
	    }
	  }
	}
	v[k]=(long)s;
      }
      tetpil=avma;return gerepile(av,tetpil,gcopy(v));
    }
  }
}

GEN
element_inv(GEN nf, GEN x)
                
/* Recoit un vecteurs de longueur N representant x dans la base 
 d'entiers et ressort son inverse sous forme d'un vecteur */
{
  long av=avma,tetpil,flx,i,N,tx=typ(x);
  GEN p1,p,z,unmod;
  
  nf=checknf(nf);N=lgef((GEN)nf[1])-3;
  if(tx<=10)
  {
    if(tx==9)
    {
      if(!gegal((GEN)x[1],(GEN)nf[1])) 
	err(talker,"not the same polynomial in a number field operation");
    }
    if(tx==10) x=gmodulcp(x,(GEN)nf[1]);
    p1=ginv(x);tetpil=avma;
    return gerepile(av,tetpil,algtobasis(nf,p1));
  }
  if(isnfscalar(x))
  {
    z=cgetg(N+1,18);z[1]=(long)ginv((GEN)x[1]);
    for(i=2;i<=N;i++) z[i]=lcopy((GEN)x[i]);
    return z;
  }
  flx=1;for(i=1;(i<=N)&&flx;i++)
  {
    flx=(typ((GEN)x[i])!=3);if(!flx) p=(GEN)((GEN)x[i])[1];
  }
  if(!flx) x=lift(x);
  p1=lift(gdiv(gmodulcp(polun[varn((GEN)nf[1])],(GEN)nf[1]),gmodulcp(gmul((GEN)nf[7],x),(GEN)nf[1])));
  z=cgetg(N+1,18);
  for(i=1;i<=N;i++) z[i]=(long)truecoeff(p1,i-1);
  if(flx)
  {tetpil=avma;return gerepile(av,tetpil,gmul((GEN)nf[8],z));}
  else
  {
    p1=gmul((GEN)nf[8],z);unmod=gmodulcp(gun,p);tetpil=avma;
    return gerepile(av,tetpil,gmul(unmod,p1));
  }
}

GEN
element_div(GEN nf, GEN x, GEN y)
                
/* Recoit deux vecteurs de longueur N representant x et y dans la base 
 d'entiers et ressort leur quotient sous forme d'un vecteur */
{
  long av=avma,tetpil,flx,fly,i,N,tx=typ(x),ty=typ(y);
  GEN p1,p,z,unmod;

  nf=checknf(nf);N=lgef((GEN)nf[1])-3;
  if(tx==9)
  {
    if(!gegal((GEN)x[1],(GEN)nf[1]))
      err(talker,"not the same polynomial in a number field operation");
  }
  if(ty==9)
  {
    if(!gegal((GEN)y[1],(GEN)nf[1]))
      err(talker,"not the same polynomial in a number field operation");
  }
  if(tx==10) x=gmodulcp(x,(GEN)nf[1]);
  if(ty==10) y=gmodulcp(y,(GEN)nf[1]);
  if(tx<=10)
  {
    if(ty<=10) p1=gdiv(x,y);
    else
    {
      if(ty!=18) err(talker,"incorrect type in nfdiv");
      p1=gdiv(x,gmodulcp(gmul((GEN)nf[7],y),(GEN)nf[1]));
    }
    tetpil=avma;return gerepile(av,tetpil,algtobasis(nf,p1));
  }
  else
  {
    if(ty<=10)
    {
      if(tx!=18) err(talker,"incorrect type in nfdiv");
      p1=gdiv(gmodulcp(gmul((GEN)nf[7],x),(GEN)nf[1]),y);
      tetpil=avma;return gerepile(av,tetpil,algtobasis(nf,p1));
    }
    else
    {
      if(isnfscalar(y)) return gdiv(x,(GEN)y[1]);
      if(isnfscalar(x))
      {
	p1=element_inv(nf,y);tetpil=avma;
	return gerepile(av,tetpil,gmul((GEN)x[1],p1));
      }
      flx=1;for(i=1;(i<=N)&&flx;i++)
      {
	flx=(typ((GEN)x[i])!=3);if(!flx) p=(GEN)((GEN)x[i])[1];
      }
      if(!flx) x=lift(x);
      fly=1;for(i=1;(i<=N)&&fly;i++)
      {
	fly=(typ((GEN)y[i])!=3);if(!fly) p=(GEN)((GEN)y[i])[1];
      }
      if(!fly) y=lift(y);
      p1=lift(gdiv(gmodulcp(gmul((GEN)nf[7],x),(GEN)nf[1]),gmodulcp(gmul((GEN)nf[7],y),(GEN)nf[1])));
      z=cgetg(N+1,18);
      for(i=1;i<=N;i++) z[i]=(long)truecoeff(p1,i-1);
      if(flx&&fly)
      {tetpil=avma;return gerepile(av,tetpil,gmul((GEN)nf[8],z));}
      else
      {
	p1=gmul((GEN)nf[8],z);unmod=gmodulcp(gun,p);tetpil=avma;
	return gerepile(av,tetpil,gmul(unmod,p1));
      }
    }
  }
}

GEN
element_muli(GEN nf, GEN x, GEN y)
                
/* Recoit deux vecteurs de longueur N representant x et y dans la base 
 d'entiers (uniquement coeff entiers) et ressort leur produit sous forme
 d'un vecteur */

{
  long av=avma,tetpil,i,j,k,N=lgef((GEN)nf[1])-3;
  GEN s,v,c;

  v=cgetg(N+1,18);
  for(k=1;k<=N;k++)
  {
    s=gzero;
    for(i=1;i<=N;i++)
    {
      c=(GEN)coeff((GEN)nf[9],k,(i-1)*N+i);
      if(signe(c))
      {
	if(gcmp1(c)) s=addii(s,mulii((GEN)x[i],(GEN)y[i]));
	else s=addii(s,mulii(mulii((GEN)x[i],(GEN)y[i]),c));
      }
      for(j=i+1;j<=N;j++)
      {
	c=(GEN)coeff((GEN)nf[9],k,(i-1)*N+j);
	if(signe(c))
	{
	  if(gcmp1(c)) s=addii(s,addii(mulii((GEN)x[i],(GEN)y[j]),mulii((GEN)x[j],(GEN)y[i])));
	  else s=addii(s,mulii(addii(mulii((GEN)x[i],(GEN)y[j]),mulii((GEN)x[j],(GEN)y[i])),c));
	}
      }
    }
    v[k]=(long)s;
  }
  tetpil=avma;return gerepile(av,tetpil,gcopy(v));
}

GEN
element_mulh(GEN nf, long limi, long limj, GEN x, GEN y)

/* Recoit deux vecteurs de longueur N representant x et y dans la base 
 d'entiers (uniquement coeff entiers) et ressort leur produit sous forme d'un
 vecteur */

{
  long av=avma,tetpil,i,j,k,N=lgef((GEN)nf[1])-3;
  GEN s,v,c;

  if(limi<limj) {i=limi;limi=limj;limj=i;s=x;x=y;y=s;}
  v=cgetg(N+1,18);
  for(k=1;k<=N;k++)
  {
    s=gzero;
    for(i=1;i<=limj;i++)
    {
      c=(GEN)coeff((GEN)nf[9],k,(i-1)*N+i);
      if(signe(c))
      {
	if(gcmp1(c)) s=addii(s,mulii((GEN)x[i],(GEN)y[i]));
	else s=addii(s,mulii(mulii((GEN)x[i],(GEN)y[i]),c));
      }
      for(j=i+1;j<=limj;j++)
      {
	c=(GEN)coeff((GEN)nf[9],k,(i-1)*N+j);
	if(signe(c))
	{
	  if(gcmp1(c)) s=addii(s,addii(mulii((GEN)x[i],(GEN)y[j]),mulii((GEN)x[j],(GEN)y[i])));
	  else s=addii(s,mulii(addii(mulii((GEN)x[i],(GEN)y[j]),mulii((GEN)x[j],(GEN)y[i])),c));
	}
      }
    }
    for(i=limj+1;i<=limi;i++)
    {
      for(j=1;j<=limj;j++)
      {
	c=(GEN)coeff((GEN)nf[9],k,(i-1)*N+j);
	if(signe(c))
	{
	  if(gcmp1(c)) s=addii(s,mulii((GEN)x[i],(GEN)y[j]));
	  else s=addii(s,mulii(mulii((GEN)x[i],(GEN)y[j]),c));
	}
      }
    }
    v[k]=(long)s;
  }
  tetpil=avma;return gerepile(av,tetpil,gcopy(v));
}

GEN
element_sqr(GEN nf, GEN x)
              
/* Recoit un vecteurs de longueur N representant x dans la base 
 d'entiers et ressort son carre sous forme d'un
 vecteur */

{
  long av=avma,tetpil,i,j,k,N=lgef((GEN)nf[1])-3;
  GEN s,v,c;

  if(isnfscalar(x))
  {
    s=cgetg(N+1,18);s[1]=(long)gsqr((GEN)x[1]);
    for(i=2;i<=N;i++) s[i]=lcopy((GEN)x[i]);
    return s;
  }
  v=cgetg(N+1,18);
  for(k=1;k<=N;k++)
  {
    s=gzero;
    for(i=1;i<=N;i++)
    {
      c=(GEN)coeff((GEN)nf[9],k,(i-1)*N+i);
      if(signe(c))
      {
	if(gcmp1(c)) s=gadd(s,gmul((GEN)x[i],(GEN)x[i]));
	else s=gadd(s,gmul(gmul((GEN)x[i],(GEN)x[i]),c));
      }
      for(j=i+1;j<=N;j++)
      {
	c=(GEN)coeff((GEN)nf[9],k,(i-1)*N+j);
	if(signe(c))
	{
	  if(gcmp1(c)) s=gadd(s,gmul2n(gmul((GEN)x[i],(GEN)x[j]),1));
	  else s=gadd(s,gmul(gmul((GEN)x[i],(GEN)x[j]),shifti(c,1)));
	}
      }
    }
    v[k]=(long)s;
  }
  tetpil=avma;return gerepile(av,tetpil,gcopy(v));
}

GEN
element_pow(GEN nf, GEN x, GEN k)
                
/* Calcule x^k dans le corps de nombres nf */

{
  long i,f,s,av=avma,tetpil,N;
  GEN k1,y,z;

  nf=checknf(nf);N=lgef((GEN)nf[1])-3;
  if(typ(x)!=18) x=algtobasis(nf,x);
  if(typ(k)!=1) err(talker,"not an integer exponent in nfpow");
  if(isnfscalar(x))
  {
    z=cgetg(N+1,18);z[1]=(long)gpui((GEN)x[1],k,0);
    for(i=2;i<=N;i++) z[i]=lcopy((GEN)x[i]);
    return z;
  }
  s=signe(k);k1=(s>=0)?k:negi(k);z=x;f=1;y=cgetg(N+1,18);
  for(i=2;i<=N;i++) y[i]=zero;
  y[1]=un;
  while(f)
  {
    if(mpodd(k1)) y=element_mul(nf,z,y);
    k1=shifti(k1,-1);f=signe(k1);
    if(f) z=element_sqr(nf,z);
  }
  tetpil=avma;return gerepile(av,tetpil,(s>=0)?gcopy(y):element_inv(nf,y));
}

GEN
element_mulid(GEN nf, GEN x, long i)
                
/* Recoit un vecteur x de longueur N representant x dans la base d'entiers
(peut etre modulo p) et ressort x.w_i sous forme d'un
 vecteur */

{
  long av=avma,tetpil,j,k,N=lgef((GEN)nf[1])-3;
  GEN s,v,c,p1;

  v=cgetg(N+1,18);
  for(k=1;k<=N;k++)
  {
    s=gzero;
    for(j=1;j<=N;j++)
    {
      c=gcoeff((GEN)nf[9],k,(i-1)*N+j);p1=(GEN)x[j];
      if((signe(c))&&(!gcmp0(p1)))
	s=gcmp1(c)?gadd(s,p1):gadd(s,gmul(p1,c));
    }
    v[k]=(long)s;
  }
  tetpil=avma;return gerepile(av,tetpil,gcopy(v));
}


long
element_val(GEN nf, GEN x, GEN vp)
{
  long av=avma,N,w,bo,i,vd,v,e=itos((GEN)vp[3]);
  GEN denx,p1,p=(GEN)vp[1],bp,d,r;

  nf=checknf(nf);
  if((typ(vp)!=17)||(lg(vp)!=6)) err(idealer3);
  N=lgef((GEN)nf[1])-3;
  switch(typ(x))
  {
    case 1: case 4: case 5:
      return ggval(x,p)*e;
    case 10: d=gabs(subres(x,(GEN)nf[1]),0);
      x=(GEN)principalideal(nf,x)[1];break;
    case 9: d=gabs(subres((GEN)x[2],(GEN)nf[1]),0);
      x=(GEN)principalideal(nf,x)[1];break;
    case 18:
      if(lg(x)==N+1)
      {d=gabs(subres(gmul((GEN)nf[7],x),(GEN)nf[1]),0);break;}
    default: err(talker,"incorrect type in element_val");
  }
  if(isnfscalar(x)) return ggval((GEN)x[1],p)*e;
  denx=denom(x);
  if(!gcmp1(denx)) {x=gmul(denx,x);vd=ggval(denx,p);} else vd=0;
  v=ggval(d,p)+N*vd;
  if(!v) return -vd*e;
  bo=0;w=0;bp=(GEN)vp[5];
  do
  {
    x=element_muli(nf,x,bp);
    if(divise((GEN)x[N],p))
    {
      for(i=1;i<=N;i++)
      {
	p1=dvmdii((GEN)x[i],p,&r);
	if(signe(r)) goto labelelv; else x[i]=(long)p1;
      }
      w++;
    }
    else bo=1;
  }
  while((bo==0)&&(w<v));
  labelelv:
  avma=av;return w-vd*e;
}

long
element_val2(GEN nf, GEN x, GEN d, GEN vp)
/* a usage interne, pas de verifs. */
{
  long av=avma,N=lgef((GEN)nf[1])-3,w,bo,i,v;
  GEN p1,p=(GEN)vp[1],bp,r;

  v=ggval(d,p);
  if(!v) return 0;
  bo=0;w=0;bp=(GEN)vp[5];
  do
  {
    x=element_muli(nf,x,bp);
    if(divise((GEN)x[N],p))
    {
      for(i=1;i<=N;i++)
      {
	p1=dvmdii((GEN)x[i],p,&r);
	if(signe(r)) goto labelelv2; else x[i]=(long)p1;
      }
      w++;
    }
    else bo=1;
  }
  while((bo==0)&&(w<v));
  labelelv2:
  avma=av;return w;
}

GEN
basistoalg(GEN nf, GEN x)
{
  long tx=typ(x),lx=lg(x),av=avma,tetpil,i;
  GEN p1,z;

  nf=checknf(nf);  
  switch(tx)
  {
    case 17: case 18:
      for(i=1;(i<lx)&&(typ((GEN)x[i])<17);i++);
      if(i==lx)
      {
	if(tx==17) x=gtrans(x);
	tetpil=avma;p1=gmul((GEN)nf[7],x);
	return gerepile(av,tetpil,gmodulcp(p1,(GEN)nf[1]));
      }
    case 19:
      z=cgetg(lx,tx);for(i=1;i<lx;i++) z[i]=(long)basistoalg(nf,(GEN)x[i]);
      return z;
    default: z=cgetg(3,19);z[1]=lcopy((GEN)nf[1]);
      z[2]=lmul(x,polun[varn((GEN)nf[1])]);return z;
  }
}

GEN
algtobasis(GEN nf, GEN x)
{
  long tx=typ(x),lx=lg(x),av=avma,tetpil,i,v,N;
  GEN y,z;
  
  nf=checknf(nf);  
  switch(tx)
  {
    case 17: case 18: case 19:
      z=cgetg(lx,tx);for(i=1;i<lx;i++) z[i]=(long)algtobasis(nf,(GEN)x[i]);
      return z;
    case 9:
      y=(GEN)x[1];v=varn(y);setvarn(y,varn((GEN)nf[1]));
      if(!gegal((GEN)nf[1],y))
      {
	setvarn(y,v);err(talker,"not the same number field in algtobasis");
      }
      x=(GEN)x[2];
    case 10: N=lgef((GEN)nf[1])-3;
      if((tx==10)&&((lgef(x)-3)>=N)) x=gmod(x,(GEN)nf[1]);
      z=cgetg(N+1,18);for(i=1;i<=N;i++) z[i]=(long)truecoeff(x,i-1);
      tetpil=avma;return gerepile(av,tetpil,gmul((GEN)nf[8],z));
    default: N=lgef((GEN)nf[1])-3;
      z=cgetg(N+1,18);for(i=2;i<=N;i++) z[i]=zero;z[1]=lcopy(x);
      return z;
  }
}

GEN
nfdiveuc(GEN nf, GEN a, GEN b)
/* Given a and b in nf, gives an algebraic integer y in nf such that a-b.y
   is "small" */
{
  long av=avma,tetpil;
  GEN p1;
  
  p1=element_div(nf,a,b);tetpil=avma;
  return gerepile(av,tetpil,ground(p1));
}

GEN
nfmod(GEN nf, GEN a, GEN b)
/* Given a and b in nf, gives a "small" algebraic integer r in nf
   of the form a-b.y is "small" */
{
  long av=avma,tetpil;
  GEN p1;
  
  p1=gneg(element_mul(nf,b,ground(element_div(nf,a,b))));tetpil=avma;
  return gerepile(av,tetpil,gadd(a,p1));
}

GEN
nfdivres(GEN nf, GEN a, GEN b)
/* Given a and b in nf, gives a two-component vector [y,r] in nf such
   that r=a-b.y is "small" */
{
  long av=avma,tetpil;
  GEN p1,y,z;
  
  p1=gneg(element_mul(nf,b,y=ground(element_div(nf,a,b))));tetpil=avma;
  z=cgetg(3,17);z[1]=lcopy(y);z[2]=(long)gadd(a,p1);
  return gerepile(av,tetpil,z);
}



/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~									~*/
/*~			      (Z_K/I)^*					~*/
/*~									~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

GEN
zsigne(GEN nf,GEN alpha,GEN arch,long nba)
{
/* retourne le vecteur des signatures de alpha, a coefficients modulo 2 */
  long i,i1,va;
  GEN vecsign,zeromod2,unmod2,rac;

  if(typ(alpha)==18) alpha=gmul((GEN)nf[7],alpha);
  if(typ(alpha)==9) alpha=(GEN)alpha[2];
  vecsign=cgetg(nba+1,18);
  zeromod2=gmodulcp(gzero,gdeux);unmod2=gmodulcp(gun,gdeux);
  rac=(GEN)nf[6];va=varn((GEN)nf[1]);
  for(i1=0,i=1;i<lg(arch);i++)
    if(signe((GEN)arch[i]))
      vecsign[++i1]=(signe(gsubst(alpha,va,(GEN)rac[i]))>0)?(long)zeromod2:(long)unmod2;
  return vecsign;
}

GEN
nfreducemodideal(GEN nf,GEN x,GEN ideal)
/* a usage interne...Reduction de x modulo l'ideal */
{
  long N,i;
  GEN y,p1,q;
    
  N=lg(x)-1;y=gcopy(x);
  ideal=idealhermite(nf,ideal);
  for(i=N;i>=1;i--)
  {
    p1=gcoeff(ideal,i,i);q=gdivround((GEN)y[i],p1);
    if(signe(q)) y=gsub(y,gmul(q,(GEN)ideal[i]));
  }
  return y;
}

GEN
columnreducemodmatrix(GEN x,GEN y)
/* a usage interne...Reduction de la colonne x modulo la matrice y en HNF */
{
  long N,i;
  GEN z,p1,q;
    
  N=lg(x)-1;z=gcopy(x);
  for(i=N;i>=1;i--)
  {
    p1=gcoeff(y,i,i);q=gdivround((GEN)z[i],p1);
    if(signe(q)) z=gsub(z,gmul(q,(GEN)y[i]));
  }
  return z;
}

GEN
lllreducemodmatrix(GEN x,GEN y)
/* a usage interne...Reduction de la colonne x modulo la matrice y inversible
   utilisant LLL */
{
  long av=avma,tetpil;
  GEN p1;

  y=gmul(y,lllint(y));p1=gmul(y,ground(gauss(y,x)));
  tetpil=avma;return gerepile(av,tetpil,gsub(x,p1));
}
  
GEN
reducemodmatrix(GEN x, GEN y)
/* a usage interne...Reduction de la matrice x modulo la matrice y */
{
  long i,k;
  GEN z;

  y=hnf(y);
  k=lg(x)-1;z=cgetg(k+1,19);
  for(i=1;i<=k;i++) z[i]=(long)columnreducemodmatrix((GEN)x[i],y);
  return z;
}

GEN
nfreducemodidele(GEN nf,GEN g,GEN idele,GEN structarch)
{
/* un element g etant donne, calcule un element congru a g modulo l'ideal
   ideal et de meme signature aux places de arch */
  
  long i,nba;
  GEN p1,p2,ideal,arch,generator;

  if((typ(idele)==17)&&(lg(idele)==3))
  {
    ideal=(GEN)idele[1];arch=(GEN)idele[2];
    p1=nfreducemodideal(nf,g,ideal);
    nba=lg((GEN)structarch[1])-1;generator=(GEN)structarch[2];
    p2=lift(gmul((GEN)structarch[3],gadd(zsigne(nf,p1,arch,nba),zsigne(nf,g,arch,nba))));
    for(i=1;i<=nba;i++) if(signe((GEN)p2[i])) p1=element_mul(nf,p1,(GEN)generator[i]);
    if(gcmp(gnorml2(p1),gnorml2(g))>0) return g;else return p1;
  }
  else return nfreducemodideal(nf,g,idele);
}

GEN
element_powmodideal(GEN nf,GEN x,GEN k,GEN ideal)
{
  long i,f,N;
  GEN k1,y,z;

  N=lgef((GEN)nf[1])-3;k1=k;z=x;f=1;y=cgetg(N+1,18);
  y[1]=un;for(i=2;i<=N;i++) y[i]=zero;
  while(f)
  {
    if(mpodd(k1)) y=element_mulmodideal(nf,z,y,ideal);
    k1=shifti(k1,-1);f=signe(k1);
    if(f) z=element_sqrmodideal(nf,z,ideal);
  }
  return y;
}

GEN
element_powmodidele(GEN nf,GEN x,GEN k,GEN idele,GEN structarch)
{
  long i,f,N;
  GEN k1,y,z;

  N=lgef((GEN)nf[1])-3;k1=k;z=x;f=1;y=cgetg(N+1,18);
  y[1]=un;for(i=2;i<=N;i++) y[i]=zero;
  while(f)
  {
    if(mpodd(k1)) y=element_mulmodidele(nf,z,y,idele,structarch);
    k1=shifti(k1,-1);f=signe(k1);
    if(f) z=element_sqrmodidele(nf,z,idele,structarch);
  }
  return y;
}

GEN
zidealij(GEN x, GEN y)
/* etant donnes deux ideaux entiers x et y en HNF tels que x|y|x^2, trouve le
   quotient (1+x)/(1+y) [[cyc],[gen],ux^-1]. A usage interne, pas de
   verifs. */
{
  long av=avma,tetpil,j,N,c;
  GEN p1,p2,p3,p4,d,z,x1;

  x1=ginv(x);p1=gmul(x1,y);p2=smith2(p1);
  p3=gmul(x,reducemodmatrix(ginv((GEN)p2[1]),p1));N=lg(p3)-1;
  for(j=1;j<=N;j++) coeff(p3,1,j)=(long)addsi(1,gcoeff(p3,1,j));
  p4=smithclean(p2);d=(GEN)p4[3];
  c=lg(d)-1;tetpil=avma;z=cgetg(4,17);
  p1=cgetg(c+1,17);p2=cgetg(c+1,17);z[1]=(long)p1;z[2]=(long)p2;
  for(j=1;j<=c;j++) {p1[j]=lcopy(gcoeff(d,j,j));p2[j]=lcopy((GEN)p3[j]);}
  z[3]=lmul((GEN)p4[1],x1);return gerepile(av,tetpil,z);
}

GEN
zconvert(GEN nf,GEN uv,GEN x,GEN arch,GEN structarch,GEN g)
{
/* un element g generateur d'un p^k divisant x etant donne, calcule
   un element congru a g modulo p^k et a 1 modulo x/p^k et de plus
   positif aux places de arch */
  
  long i,nba;
  GEN p1,p2,generator;
  
  p1=nfreducemodideal(nf,gadd((GEN)uv[1],element_mul(nf,g,(GEN)uv[2])),x);
  nba=lg((GEN)structarch[1])-1;generator=(GEN)structarch[2];
  p2=lift(zsigne(nf,p1,arch,nba));
  for(i=1;i<=nba;i++)
    if(signe((GEN)p2[i])) p1=element_mul(nf,p1,(GEN)generator[i]);
  return p1;
}

GEN
compute_prhall(GEN nf,GEN pr)
{
/* calcule le prhall associe a pr necessaire pour faire des reductions modulo
   pr meme pour des non entiers algebriques */
  
  long N;
  GEN prh,epr,betae,p1,prhall,id;

  N=lgef((GEN)nf[1])-3;id=idmat(N);
  prh=idealmul(nf,id,pr);epr=(GEN)pr[3];
  betae=gdiv(element_pow(nf,(GEN)pr[5],epr),gpui((GEN)pr[1],addis(epr,-1),0));
  p1=cgetg(2,19);p1[1]=(long)betae;
  p1=idealadd(nf,gmul((GEN)pr[1],id),idealmul(nf,p1,id));
  prhall=cgetg(3,17);prhall[1]=(long)prh;
  prhall[2]=idealaddone(nf,pr,p1)[2];
  return prhall;
}

long
nfsearch(GEN smalltable,GEN x)
/* rend l'indice de x dans la table smalltable, sinon 0 */
{
  long l,u,i,s;

  u=lg(smalltable)-1;l=1;
  while(u>=l)
  {
    i=(l+u)>>1;s=lexcmp(x,(GEN)smalltable[i]);
    if(!s) return i;
    if(s<0) u=i-1;else l=i+1;
  }
  return 0;
}

GEN
nfshanks(GEN nf,GEN x,GEN g0,GEN pr,GEN prhall)
/* rend le plus petit entier positif n tel que g0^n=x modulo pr */
{
  long av=avma,tetpil,i,N,lbaby,fl,k;
  GEN pf1,pfqr,p1,smalltable,p2,giant,smalltable2,permtable,v,g0inv;

  N=lgef((GEN)nf[1])-3;
  pf1=addsi(-1,gpui((GEN)pr[1],(GEN)pr[4],0));pfqr=racine(pf1);
  for(i=2;(i<lg(x))&&gcmp0((GEN)x[i]);i++);
  if((i==lg(x))&&gcmp_1((GEN)x[1]))
  {
    if(cmpii((GEN)pr[1],gdeux))
    {tetpil=avma;return gerepile(av,tetpil,gmul2n(pf1,-1));}
    else {avma=av;return gzero;}
  }
  if(cmpis(pfqr,65535)>=0) err(talker,"module is too large in nfshanks");
  lbaby=itos(pfqr);
  g0inv=element_invmodpr(nf,g0,prhall);
  p1=x;smalltable=cgetg(lbaby+2,17);fl=0;
  for(i=0;(i<=lbaby)&&(!fl);i++)
  {
    p2=lift(p1);smalltable[i+1]=(long)p2;p2[1]=(long)addsi(-1,(GEN)p2[1]);fl=gcmp0(p2);
    if(!fl) fl=element_val(nf,p2,pr);p2[1]=(long)addsi(1,(GEN)p2[1]);
    if((i<lbaby)&&(!fl)) p1=element_mulmodpr(nf,p1,g0inv,prhall);
  }
  if(fl) {avma=av;return stoi(i-1);}
  permtable=indexlexsort(smalltable);
  smalltable2=cgetg(lbaby+2,17);
  for(i=1;i<=lbaby+1;i++) smalltable2[i]=smalltable[itos((GEN)permtable[i])];
  smalltable=smalltable2;
  giant=element_divmodpr(nf,x,p1,prhall);p1=giant;k=1;
  while(1)
  {
    if((i=nfsearch(smalltable,lift(p1))))
    {
      v=addii(mulsi(lbaby,stoi(k)),(GEN)permtable[i]);tetpil=avma;
      return gerepile(av,tetpil,addsi(-1,v));
    }
    p1=element_mulmodpr(nf,p1,giant,prhall);k++;
  }
}

GEN compute_class_number(GEN mit,GEN *met,GEN *u1);

GEN
dethnf(GEN mat)
{
  long i,av=avma,tetpil;
  GEN s;
  
  s=gun;for(i=1;i<lg(mat);i++) s=gmul(s,gcoeff(mat,i,i));
  tetpil=avma;return gerepile(av,tetpil,gcopy(s));
}

GEN
makeprimetoideal(GEN nf,GEN ideal,GEN prk,GEN *ptuv,GEN v)
{
  /* a usage interne...*/
  
  GEN uv;

  if(*ptuv==gzero) {uv=idealaddone(nf,prk,idealdivexact(nf,ideal,prk));*ptuv=uv;}
  else uv=*ptuv;
  return nfreducemodideal(nf,gadd((GEN)uv[1],element_mul(nf,v,(GEN)uv[2])),ideal);
}

GEN
makeprimetoidealvec(GEN nf,GEN ideal,GEN ideal1,GEN ideal2,GEN listgen)
{
  /* a usage interne...*/
  
  long i,fl,llist;
  GEN uv,y;

  llist=lg(listgen)-1;fl=0;
  y=cgetg(llist+1,17);
  for(i=1;i<=llist;i++)
  {
    if(!fl){fl=1;uv=idealaddone(nf,ideal1,ideal2);}
    y[i]=(long)nfreducemodideal(nf,gadd((GEN)uv[1],element_mul(nf,(GEN)listgen[i],(GEN)uv[2])),ideal);
  }
  return y;
}

void
checkbid(GEN bid)
{
  if((typ(bid)!=17)||(lg(bid)!=6))
    err(talker,"incorrect bigideal");
}

GEN
zprimestar(GEN nf,GEN pr,GEN ep,GEN x,GEN arch,long nba,long flag)
/* etant donne un ideal pr^ep, et un ideal entier x obligatoirement en HNF,
   calcule une liste de vecteurs a cinq composantes :
   [....,[[clh],[gen1],[gen2],[signat2],U.X^-1],...] chaque composante
   correspondant a d_i,g_i,g'_i,s_i .
   Les generateurs g_i ne sont pas premiers a x, les generateurs g'_i le sont,
   et signat2 est le vecteur horizontal des vecteurs verticaux des signatures
   des g'_i. Si flag=1 on rend gen2=gen1 */
{
  long av,N,f,k,nbp,psim,j,fl,n,m,fl2,tetpil,i,e,a,b;
  GEN prh,p,pef,pefm1,fa,list,v,prhall,p1,p2,prk,list_of_struct,uv,g0;
  GEN newgen,p3,p4,pra,prb;

  av=avma;
  prh=idealhermite(nf,pr);N=lg(prh)-1;
  f=itos((GEN)pr[4]);k=itos(ep);p=(GEN)pr[1];pef=gpuigs(p,f);
  pefm1=gaddgs(pef,-1);fa=factor(pefm1);list=(GEN)fa[1];nbp=lg(list)-1;
  v=cgetg(N+1,18);psim=itos(p);
  if(f==1){v[1]=gener(p)[2];for(j=2;j<=N;j++) v[j]=zero;}
  else
  {
    prhall=cgetg(3,17);prhall[1]=(long)prh;prhall[2]=zero;
    fl=1;
    for(n=psim;fl;n++)
    {
      m=n;
      for(i=1;i<=N;i++)
      {if(!gcmp1(gcoeff(prh,i,i))) {v[i]=lstoi(m%psim);m/=psim;}else v[i]=zero;}
      fl2=1;
      for(j=1;(j<=nbp)&&fl2;j++)
      {
	p1=lift(element_powmodpr(nf,v,divii(pefm1,(GEN)list[j]),prhall));
	p1[1]=(long)addsi(-1,(GEN)p1[1]);
	if(gcmp0(p1)||idealval(nf,p1,pr)) fl2=0;
      }
      if(fl2) fl=0;
    }
  }
  prk=idealpow(nf,pr,ep);uv=gzero;
  if(flag) g0=v;
  else g0=makeprimetoideal(nf,x,prk,&uv,v);
  list_of_struct=cgetg(2,17);
  p1=cgetg(6,17);list_of_struct[1]=(long)p1;
  p2=cgetg(2,17);p1[1]=(long)p2;p2[1]=(long)pefm1;
  p2=cgetg(2,17);p1[2]=(long)p2;p2[1]=(long)v;p1[5]=un;
  p2=cgetg(2,17);p1[3]=(long)p2;p2[1]=(long)g0;
  p2=cgetg(2,17);p1[4]=(long)p2;p2[1]=(long)zsigne(nf,g0,arch,nba);
  if(gcmp1(ep)){tetpil=avma;return gerepile(av,tetpil,gcopy(list_of_struct));}
  else
  {
    e=itos(ep);a=1;b=2;pra=idealhermite(nf,pr);prb=idealmul(nf,pra,pr);
    while(a<b)
    {
      p1=zidealij(pra,prb);newgen=gcopy((GEN)p1[2]);p3=cgetg(lg(newgen),17);
      for(i=1;i<lg(newgen);i++)
      {
	if(!flag) newgen[i]=(long)makeprimetoideal(nf,x,prk,&uv,(GEN)newgen[i]);
	p3[i]=(long)zsigne(nf,(GEN)newgen[i],arch,nba);
      }
      p4=cgetg(6,17);
      p4[1]=p1[1];p4[2]=p1[2];p4[3]=(long)newgen;p4[4]=(long)p3;p4[5]=p1[3];
      p2=cgetg(2,17);p2[1]=(long)p4;list_of_struct=concat(list_of_struct,p2);
      a=b;b=min(e,b<<1);
      if(a!=b)
      {
	pra=prb;
	if(b==(a<<1)) prb=idealmul(nf,prb,prb);
	else prb=prk;
      }
    }
    tetpil=avma;return gerepile(av,tetpil,gcopy(list_of_struct));
  }
}

GEN
zarchstar(GEN nf,GEN x,GEN arch,long nba)
{
/* un ideal x etant donne, calcule des elements de 1+x ayant une matrice
   de signatures inversible */

  
  long av=avma,av1,tetpil,N,i,j,fl,r,rr,limr,k,kk,rankinit,ranknew;
  GEN p1,p2,y,bas,v2,v3,vecsign,genarch,alpha,lambda;

  if(!nba)
  {
    y=cgetg(4,17);p1=cgetg(1,17);y[1]=(long)p1;
    p2=cgetg(1,17);y[2]=(long)p2;p2=cgetg(1,19);y[3]=(long)p2;return y;
  }
  x=gmul(x,lllintpartial(x));
  N=lgef((GEN)nf[1])-3;
  v2=cgetg(1,19);
  r=1;rr=3;bas=gmul((GEN)nf[7],x);rankinit=0;j=0;fl=1;
  lambda=cgeti(N+1);genarch=cgetg(nba+1,17);
  /* dans cette boucle on va chercher des elements pour que la matrice
     des signatures modulo 2 soit inversible */
  while(fl)
  {
    limr=(N==1)?rr:(itos(gmin(stoi(BIGINT),gpuigs(stoi(rr),N)))-1)>>1;
    for(k=rr;(k<=limr)&&fl;k++)
    {
      if(N==1) lambda[1]=-2;
      else
      {kk=k;for(i=1;i<=N;i++){lambda[i]=(kk+r)%rr-r;kk/=rr;}}
      av1=avma;
      alpha=gun;for(i=1;i<=N;i++) alpha=gadd(alpha,gmulsg(lambda[i],(GEN)bas[i]));
      vecsign=zsigne(nf,alpha,arch,nba);
      if(lg(v2)>1) v3=concat(v2,gtomat(vecsign));
      else v3=gtomat(vecsign);ranknew=rank(v3);
      if(ranknew>rankinit)
      {
	v2=v3;rankinit++;fl=(rankinit<nba);
	genarch[++j]=(long)algtobasis(nf,alpha);
      }
      else avma=av1;
    }
    if(fl) {r++;rr+=2;}
  }
  if(rankinit<nba) err(talker,"bug in zarchstar");
  v3=lift(ginv(v2));
  tetpil=avma;y=cgetg(4,17);p1=cgetg(nba+1,17);y[1]=(long)p1;
  for(i=1;i<=nba;i++) p1[i]=(long)gdeux;
  y[2]=lcopy(genarch);y[3]=lcopy(v3);return gerepile(av,tetpil,y);
}

GEN
zinternallog(GEN nf,GEN set_of_list,long nbgen,GEN arch,GEN fa,GEN a,long indice)
/* A usage interne : retourne la decomposition de a sur les nbgen generateurs
   successifs contenus dans set_of_list et si indice !=0 on ne fait ce calcul
   que pour l'ideal premier correspondant a cet indice en mettant a 0 les
   autres composantes */
{
  long av=avma,tetpil,nbp,cp,k,e,j,i,nba,fl;
  GEN list,ep,y,ainit,list_of_struct,primetop,g0brut,pr,p,prhall,structarch,pk1,pk;
  GEN prk,p1,p2,b,p3,p4,p6,psigne;

  list=(GEN)fa[1];ep=(GEN)fa[2];nbp=lg(list)-1;cp=0;
  if(typ(a)<=10) a=algtobasis(nf,a);
  ainit=gcopy(a);
  structarch=(GEN)set_of_list[nbp+1];nba=lg((GEN)structarch[1])-1;
  psigne=zsigne(nf,ainit,arch,nba);
  y=cgetg(nbgen-nba+1,18);
  for(k=1;k<=nbp;k++)
  {
    fl=(!indice)||(indice==k);
    a=ainit;
    list_of_struct=(GEN)set_of_list[k];
    if(fl)
    {
      primetop=(GEN)list_of_struct[1];
      g0brut=(GEN)((GEN)primetop[2])[1];
      pr=(GEN)list[k];p=(GEN)pr[1];e=itos((GEN)ep[k]);prhall=compute_prhall(nf,pr);
      y[++cp]=(long)nfshanks(nf,a,g0brut,pr,prhall);
      if(mpodd((GEN)y[cp])) psigne=gadd(psigne,(GEN)((GEN)primetop[4])[1]);
    }
    else y[++cp]=zero;
    if(fl)
    {
      pk1=gpuigs(p,e-1);pk=gmul(pk1,p);prk=idealpow(nf,pr,(GEN)ep[k]);
      p2=element_powmodideal(nf,g0brut,(GEN)y[cp],prk);
      a=element_divmodideal(nf,a,p2,prk);
    }
    for(j=2;j<lg(list_of_struct);j++)
    {
      p1=(GEN)list_of_struct[j];
      if(fl)
      {
	b=gcopy(a);b[1]=(long)addsi(-1,(GEN)b[1]);
	p2=gmul((GEN)p1[5],b);p3=(GEN)p1[1];p4=(GEN)p1[2];
	if(lg(p2)!=lg(p3)) err(talker,"bug in zinternallog");
	for(i=1;i<lg(p3);i++)
	{
	  y[++cp]=(long)negi(p6=modii(negi((GEN)p2[i]),(GEN)p3[i]));
	  if(mpodd((GEN)y[cp])) psigne=gadd(psigne,(GEN)((GEN)p1[4])[i]);
	  a=element_mulmodideal(nf,element_powmodideal(nf,(GEN)p4[i],p6,prk),a,prk);
	}
      }
      else for(i=1;i<lg((GEN)p1[1]);i++) y[++cp]=zero;
    }
  }
  p1=lift(gmul((GEN)structarch[3],psigne));tetpil=avma;
  return gerepile(av,tetpil,concat(y,p1));
}

GEN
zidealstarinitall(GEN nf, GEN ideal,long flun)
/* Calcule [[ideal,arch],[h,[cyc],[gen]],idealfact,[liste],U]
   gen n'est pas calcule si flun=0 */
{
  long av=avma,tetpil,i,j,k,nba,nbp,N,fl,c,s,R1,nbgen,cp,jj;
  GEN p1,p2,p3,p3plus,p3moins,p4,p5,p6,y,h,clh,met,u1,basecl,mot,generator;
  GEN fa,fa2,list,ep,x,arch,allgenerator,list_of_struct,structarch,u1u2;
  GEN unnf,ideal2;
  
  nf=checknf(nf);N=lgef((GEN)nf[1])-3;R1=itos((GEN)((GEN)nf[2])[1]);
  if((typ(ideal)==17)&&(lg(ideal)==3))
  {
    fl=1;ideal2=cgetg(3,17);x=idealhermite(nf,(GEN)ideal[1]);ideal2[1]=(long)x;
    arch=(GEN)ideal[2];nba=0;
    if(((typ(arch)!=17)&&(typ(arch)!=18))||(lg(arch)!=1+R1))
      err(talker,"incorrect archimedean component in zidealstarinit");
    for(i=1;i<=R1;i++) if(signe((GEN)arch[i])) nba++;
    ideal2[2]=lcopy(arch);ideal=ideal2;
  }
  else
  {
    fl=0;x=idealhermite(nf,ideal);
    arch=cgetg(R1+1,17);for(i=1;i<=R1;i++) arch[i]=zero;nba=0;
    ideal=cgetg(3,17);ideal[1]=(long)x;ideal[2]=(long)arch;
  }
  if(!gcmp1(denom(x))) err(talker,"zidealstarinit needs an integral ideal");
  y=cgetg(6,17);y[1]=(long)ideal;
  fa=idealfactor(nf,x);list=(GEN)fa[1];ep=(GEN)fa[2];nbp=lg(list)-1;
  y[3]=(long)fa;fa2=cgetg(nbp+2,17);y[4]=(long)fa2;
  structarch=zarchstar(nf,x,arch,nba);allgenerator=cgetg(1,17);
  for(i=1;i<=nbp;i++)
  {
    p1=zprimestar(nf,(GEN)list[i],(GEN)ep[i],x,arch,nba,nbp==1);
    fa2[i]=(long)p1;
    for(j=1;j<lg(p1);j++)
      allgenerator=concat(allgenerator,(GEN)((GEN)p1[j])[3]);
  }
  fa2[nbp+1]=(long)structarch;
  allgenerator=concat(allgenerator,(GEN)structarch[2]);
  nbgen=lg(allgenerator)-1;
  h=cgetg(nbgen+1,19);cp=0;
  for(k=1;k<lg(fa2)-1;k++)
  {
    list_of_struct=(GEN)fa2[k];
    for(j=1;j<lg(list_of_struct);j++)
    {
      p1=(GEN)list_of_struct[j];p2=(GEN)p1[1];p3=(GEN)p1[3];
      for(jj=1;jj<lg((GEN)p1[1]);jj++)
      {
	p5=element_powmodidele(nf,(GEN)p3[jj],(GEN)p2[jj],ideal,structarch);
	h[++cp]=(long)gneg(zinternallog(nf,fa2,nbgen,arch,fa,p5,k));
	coeff(h,cp,cp)=((GEN)p1[1])[jj];
      }
    }
  }
  p1=cgetg(nbgen+1,18);for(i=1;i<=nbgen;i++) p1[i]=zero;
  for(j=1;j<=nba;j++) {h[++cp]=lcopy(p1);coeff(h,cp,cp)=deux;}
  if(cp!=nbgen) err(talker,"bug in zidealstarinit");
  if(nbgen)
  {
    u1u2=smith2(h);u1=reducemodmatrix(ginv((GEN)u1u2[1]),h);
    u1u2=smithclean(u1u2);
    met=(GEN)u1u2[3];y[5]=(long)(GEN)u1u2[1];
    clh=dethnf(met);c=lg(met)-1;
  }
  else {clh=gun;met=cgetg(1,19);u1=cgetg(1,19);y[5]=(long)u1;c=0;}
  if(flun)
  {
    basecl=cgetg(c+1,17);
    unnf=cgetg(N+1,18);unnf[1]=un;for(i=2;i<=N;i++) unnf[i]=zero;
    for(j=1;j<=c;j++)
    {
      p3plus=unnf;p3moins=unnf;
      for(i=1;i<=nbgen;i++)
      {
	p1=gcoeff(u1,i,j);s=signe(p1);
	if(s)
	{
	  if(s>0)
	    p3plus=element_mulmodidele(nf,p3plus,(GEN)element_powmodidele(nf,(GEN)allgenerator[i],p1,ideal,structarch),ideal,structarch);
	  else
	    p3moins=element_mulmodidele(nf,p3moins,(GEN)element_powmodidele(nf,(GEN)allgenerator[i],negi(p1),ideal,structarch),ideal,structarch);
	}
      }
      if(nbp)
      {
	p5=idealhermite(nf,p3moins);
	p6=idealaddone(nf,p5,x);
	p4=element_div(nf,(GEN)p6[1],p3moins);
/* p4 est l'inverse de p3moins modulo x */
	p3=element_mulmodideal(nf,p3plus,p4,x);
      }
      else p3=unnf;
      generator=(GEN)structarch[2];
      p5=gadd(gadd(zsigne(nf,p3,arch,nba),zsigne(nf,p3plus,arch,nba)),zsigne(nf,p3moins,arch,nba));
      p2=lift(gmul((GEN)structarch[3],p5));
      for(i=1;i<=nba;i++)
	if(signe((GEN)p2[i])) p3=element_mul(nf,p3,(GEN)generator[i]);
/* on a corrige pour que p3 ait la meme signature que p3plus/p3moins */
      basecl[j]=(long)p3;
    }
    p1=cgetg(4,17);p1[1]=(long)clh;
    mot=cgetg(c+1,17);for(i=1;i<=c;i++) mot[i]=coeff(met,i,i);
    p1[2]=(long)mot;p1[3]=(long)basecl;
  }
  else
  {
    p1=cgetg(3,17);p1[1]=(long)clh;
    mot=cgetg(c+1,17);for(i=1;i<=c;i++) mot[i]=coeff(met,i,i);
    p1[2]=(long)mot;
  }
  y[2]=(long)p1;
  tetpil=avma;return gerepile(av,tetpil,gcopy(y));
}

GEN
zidealstarinitgen(GEN nf, GEN ideal)
{
  return zidealstarinitall(nf,ideal,1);
}

GEN
zidealstarinit(GEN nf, GEN ideal)
{
  return zidealstarinitall(nf,ideal,0);
}

GEN
zidealstar(GEN nf, GEN ideal)
{
  long av=avma,tetpil;
  GEN y;
  
  y=zidealstarinitall(nf,ideal,1);
  tetpil=avma;return gerepile(av,tetpil,gcopy((GEN)y[2]));
}

GEN
zideallog(GEN nf, GEN x, GEN bigideal)
/* etant donne un element x eventuellement non entier, et un idele ou ideal
   entier en format donne par zidealstarinit, calcule le vecteur des
   composantes sur les generateurs donnes dans bigideal[2] */
{
  long av=avma,tetpil,nbp,sizeh,i,v,k,j,N,c,R1,nba;
  GEN fa,fa2,list,ep,y,pr,p1,p2,p3,x1,den,dinv,ideal,ideal1,cyclic,arch,id,structarch;

  nf=checknf(nf);
  checkbid(bigideal);
  N=lgef((GEN)nf[1])-3;R1=itos((GEN)((GEN)nf[2])[1]);
  ideal=(GEN)bigideal[1];
  if((typ(ideal)==17)&&(lg(ideal)==3))
  {ideal1=(GEN)ideal[1];arch=(GEN)ideal[2];}
  else {ideal1=ideal;arch=cgetg(R1+1,17);for(i=1;i<=R1;i++) arch[i]=zero;}
  fa=(GEN)bigideal[3];fa2=(GEN)bigideal[4];list=(GEN)fa[1];ep=(GEN)fa[2];
  nbp=lg(list)-1;id=idmat(N);
  if(lg(fa2)!=(nbp+2)) err(talker,"not a big ideal in zideallog");
  structarch=(GEN)fa2[nbp+1];
  nba=lg((GEN)structarch[1])-1;
  sizeh=lg((GEN)bigideal[5])-1;
  if(gcmp0(x)) err(talker,"element is not coprime to ideal in zideallog");
  switch(typ(x))
  {
    case 1: case 4: case 5:
      p1=cgetg(N+1,18);p1[1]=(long)x;for(j=2;j<=N;j++) p1[j]=zero;x=p1;break;
    case 9: case 10: x=algtobasis(nf,x);break;
    case 18: break;
    default: err(talker,"not an element in zideallog");
  }
  if(lg(x)!=(N+1)) err(talker,"not an element in zideallog");
  den=denom(x);
  if(!gcmp1(den))
  {
	/* ici x n'est pas entier, mais on le suppose tel que v_p(x)=0 pour
	   tout p divisant l'ideal. Il faut ecrire x=a/b avec a et b entiers
	   premiers a l'ideal */
    k=1;
    for(i=1;i<=nbp;i++)
    {
      pr=(GEN)list[i];
      v=(ggval(den,(GEN)pr[1])*itos((GEN)pr[3]))/itos((GEN)ep[i])+1;
      if(v>k) k=v;
    }
    p2=gmul(den,id);p3=idealpow(nf,ideal1,stoi(k));
    p1=idealadd(nf,p2,p3);
    dinv=idealinv(nf,p1);
    p2=idealmullll(nf,p2,dinv);p3=idealmullll(nf,p3,dinv);
    x1=(GEN)idealaddone(nf,p2,p3)[1];
    if(gcmp0(x1)) x1=gmul(den,(GEN)id[1]);
	/* on a trouve x=a/b comme on veut, avec a=x1*x, b=x1 */
    p1=zideallog(nf,element_mul(nf,x1,x),bigideal);
    p2=zideallog(nf,x1,bigideal);
    p1=gsub(p1,p2);
    p2=(GEN)((GEN)bigideal[2])[2];c=lg(p2)-1;
    tetpil=avma;y=cgetg(c+1,18);
    for(i=1;i<=c;i++) y[i]=(long)modii((GEN)p1[i],(GEN)p2[i]);
    return gerepile(av,tetpil,y);
  }
  y=zinternallog(nf,fa2,sizeh,arch,fa,x,0);
  cyclic=(GEN)((GEN)bigideal[2])[2];
  c=lg(cyclic)-1;
  p1=gmul((GEN)bigideal[5],y); /* on applique le smith a y */
  if(lg(p1)-1!=c) err(talker,"bug in zideallog, please report");
  tetpil=avma;y=cgetg(c+1,18);
  for(i=1;i<=c;i++) y[i]=(long)modii((GEN)p1[i],(GEN)cyclic[i]);
  return gerepile(av,tetpil,y);
}

GEN
zidealstarinitjoinall(GEN nf, GEN bid1, GEN bid2, long flun)
/* Etant donnes bid1, bid2 resultats de zidealstarinit pour deux modules
   m1 et m2 premiers entre eux sans partie archimedienne, calcule le
   zidealstarinit [[ideal,arch],[h,[cyc],[gen]],idealfact,[liste],U]
   du produit */
{
  long av=avma,tetpil,i,j,nbp,N,c,s,R1,nbgen,lx1,lx2,llx1,llx2,lx,llx,fltr;
  GEN module1,module2,struct1,struct2,fact1,fact2,liste1,liste2,U1,U2;
  GEN module,liste,fact,U,cyc,ex1,ex2;
  GEN p1,p2,p3,p3plus,p3moins,p4,p5,p6,y,clh,met,u1,basecl,mot;
  GEN fa1,fa2,x,allgenerator,u1u2;
  GEN unnf;
  
  nf=checknf(nf);N=lgef((GEN)nf[1])-3;R1=itos((GEN)((GEN)nf[2])[1]);
  checkbid(bid1);checkbid(bid2);
  module1=(GEN)bid1[1];struct1=(GEN)bid1[2];fact1=(GEN)bid1[3];
  liste1=(GEN)bid1[4];U1=(GEN)bid1[5];
  module2=(GEN)bid2[1];struct2=(GEN)bid2[2];fact2=(GEN)bid2[3];
  liste2=(GEN)bid2[4];U2=(GEN)bid2[5];
  module=cgetg(3,17);module[1]=(long)idealmul(nf,(GEN)module1[1],(GEN)module2[1]);
  module[2]=(long)gadd((GEN)module1[2],(GEN)module2[2]);x=(GEN)module[1];
  if(gcmpgs(vecmax((GEN)module[2]),1)>=0)
    err(talker,"nontrivial Archimedian components in zidealstarinitjoin");
  fa1=(GEN)fact1[1];ex1=(GEN)fact1[2];fa2=(GEN)fact2[1];ex2=(GEN)fact2[2];
  fact=cgetg(3,19);fact[1]=(long)concat(fa1,fa2);fact[2]=(long)concat(ex1,ex2);
  lx1=lg(fa1);lx2=lg(fa2);fltr=0;
  for(i=1;(i<lx1)&&(!fltr);i++) fltr=isinvector(fa2,(GEN)fa1[i],lx2-1);
  if(fltr) err(talker,"noncoprime ideals in zidealstarinitjoin");
  liste1=(GEN)bid1[4];liste2=(GEN)bid2[4];lx1=lg(liste1);lx2=lg(liste2);
  lx=lx1+lx2-2;liste=cgetg(lx,17);
  for(i=1;i<lx1-1;i++) liste[i]=liste1[i];
  for(i=lx1-1;i<lx;i++) liste[i]=liste2[i-lx1+2];
  y=cgetg(6,17);y[1]=(long)module;y[3]=(long)fact;y[4]=(long)liste;
  lx1=lg(U1);lx2=lg(U2);lx=lx1+lx2-1;
  llx1=lg((GEN)struct1[2]);llx2=lg((GEN)struct2[2]);
  llx=llx1+llx2-1;nbgen=llx-1;
  if(nbgen)
  {
    cyc=diagonal(concat((GEN)struct1[2],(GEN)struct2[2]));
    u1u2=smith2(cyc);u1=(GEN)u1u2[1];
    u1u2=smithclean(u1u2);met=(GEN)u1u2[3];clh=dethnf(met);c=lg(met)-1;
    U=cgetg(lx,19);
    for(j=1;j<lx1;j++)
    {
      p1=cgetg(llx,18);p2=(GEN)U1[j];U[j]=(long)p1;
      for(i=1;i<llx1;i++) p1[i]=p2[i];
      for(;i<llx;i++) p1[i]=zero;
    }
    for(;j<lx;j++)
    {
      p1=cgetg(llx,18);p2=(GEN)U2[j-lx1+1];U[j]=(long)p1;
      for(i=1;i<llx1;i++) p1[i]=zero;
      for(;i<llx;i++) p1[i]=p2[i-llx1+1];
    }
    y[5]=lmul((GEN)u1u2[1],U);
  }
  else
  {
    clh=gun;met=cgetg(1,19);
    U=cgetg(lx,19);for(j=1;j<lx;j++) U[j]=lgetg(1,18);
    y[5]=(long)U;c=0;
  }
  if(flun)
  {
    if((lg(struct1)<=3)||(lg(struct2)<=3))
      err(talker,"please apply zidealstarinitgen and not zidealstarinit");
    u1=reducemodmatrix(ginv(u1),cyc);
    p1=makeprimetoidealvec(nf,(GEN)module[1],(GEN)module1[1],(GEN)module2[1],(GEN)struct1[3]);
    p2=makeprimetoidealvec(nf,(GEN)module[1],(GEN)module2[1],(GEN)module1[1],(GEN)struct2[3]);
    allgenerator=concat(p1,p2);
    basecl=cgetg(c+1,17);
    unnf=cgetg(N+1,18);unnf[1]=un;for(i=2;i<=N;i++) unnf[i]=zero;
    for(j=1;j<=c;j++)
    {
      p3plus=unnf;p3moins=unnf;
      for(i=1;i<=nbgen;i++)
      {
	p1=gcoeff(u1,i,j);s=signe(p1);
	if(s)
	{
	  if(s>0)
	    p3plus=element_mul(nf,p3plus,(GEN)element_pow(nf,(GEN)allgenerator[i],p1));
	  else
	    p3moins=element_mul(nf,p3moins,(GEN)element_pow(nf,(GEN)allgenerator[i],negi(p1)));
	}
      }
      if(nbp)
      {
	p5=idealhermite(nf,p3moins);
	p6=idealaddone(nf,p5,x);
	p4=element_div(nf,(GEN)p6[1],p3moins);
/* p4 est l'inverse de p3moins modulo x */
	p3=element_mulmodideal(nf,p3plus,p4,x);
      }
      else p3=unnf;
      basecl[j]=(long)p3;
    }
    p1=cgetg(4,17);p1[1]=(long)clh;
    mot=cgetg(c+1,17);for(i=1;i<=c;i++) mot[i]=coeff(met,i,i);
    p1[2]=(long)mot;p1[3]=(long)basecl;
  }
  else
  {
    p1=cgetg(3,17);p1[1]=(long)clh;
    mot=cgetg(c+1,17);for(i=1;i<=c;i++) mot[i]=coeff(met,i,i);
    p1[2]=(long)mot;
  }
  y[2]=(long)p1;
  tetpil=avma;return gerepile(av,tetpil,gcopy(y));
}

GEN
zidealstarinitjoin(GEN nf, GEN bid1, GEN bid2)
{
  return zidealstarinitjoinall(nf,bid1,bid2,0);
}

GEN
zidealstarinitjoingen(GEN nf, GEN bid1, GEN bid2)
{
  return zidealstarinitjoinall(nf,bid1,bid2,1);
}

GEN
zidealstarinitjoinarchall(GEN nf, GEN bid1, GEN arch, long nba, long flun)
/* Etant donnes bid1 resultat de zidealstarinit pour un module
   m1 sans partie archimedienne et une partie archimedienne arch, calcule
   le zidealstarinit [[ideal,arch],[h,[cyc],[gen]],idealfact,[liste],U]
   du produit */
{
  long av=avma,tetpil,i,j,nbp,N,c,s,R1,nbgen,lx1,lx,llx1,llx;
  GEN module1,struct1,fact1,liste1,U1;
  GEN module,liste,cyc;
  GEN p1,p2,p3,p3plus,p3moins,p4,p5,p6,y,clh,met,u1,basecl,mot,generator;
  GEN x,allgenerator,structarch,u1u2;
  GEN unnf;
  
  nf=checknf(nf);N=lgef((GEN)nf[1])-3;R1=itos((GEN)((GEN)nf[2])[1]);
  checkbid(bid1);
  module1=(GEN)bid1[1];struct1=(GEN)bid1[2];fact1=(GEN)bid1[3];
  liste1=(GEN)bid1[4];U1=(GEN)bid1[5];
  x=(GEN)module1[1];
  structarch=zarchstar(nf,x,arch,nba);
  module=cgetg(3,17);module[1]=module1[1];module[2]=(long)arch;
  if(gcmpgs(vecmax((GEN)module1[2]),1)>=0)
    err(talker,"nontrivial Archimedian components in zidealstarinitjoinarchall");
  liste1=(GEN)bid1[4];lx=lg(liste1);  
  liste=cgetg(lx,17);
  for(i=1;i<lx-1;i++) liste[i]=liste1[i];
  liste[lx-1]=(long)structarch;
  lx1=lg(U1);lx=lx1+nba;llx1=lg((GEN)struct1[2]);llx=llx1+nba;nbgen=llx-1;
  y=cgetg(6,17);y[1]=(long)module;y[3]=(long)fact1;y[4]=(long)liste;
  if(nbgen)
  {
    cyc=diagonal(concat((GEN)struct1[2],(GEN)structarch[1]));
    u1u2=smith2(cyc);u1=(GEN)u1u2[1];
    u1u2=smithclean(u1u2);met=(GEN)u1u2[3];clh=dethnf(met);c=lg(met)-1;
    y[5]=u1u2[1];
  }
  else {clh=gun;met=cgetg(1,19);u1=cgetg(1,19);y[5]=(long)u1;c=0;}
  if(flun)
  {
    if(lg(struct1)<=3)
      err(talker,"please apply zidealstarinitgen and not zidealstarinit");
    u1=reducemodmatrix(ginv(u1),cyc);
    allgenerator=concat((GEN)struct1[3],(GEN)structarch[2]);
    basecl=cgetg(c+1,17);
    unnf=cgetg(N+1,18);unnf[1]=un;for(i=2;i<=N;i++) unnf[i]=zero;
    for(j=1;j<=c;j++)
    {
      p3plus=unnf;p3moins=unnf;
      for(i=1;i<=nbgen;i++)
      {
	p1=gcoeff(u1,i,j);s=signe(p1);
	if(s)
	{
	  if(s>0)
	    p3plus=element_mul(nf,p3plus,(GEN)element_pow(nf,(GEN)allgenerator[i],p1));
	  else
	    p3moins=element_mul(nf,p3moins,(GEN)element_pow(nf,(GEN)allgenerator[i],negi(p1)));
	}
      }
      if(nbp)
      {
	p5=idealhermite(nf,p3moins);
	p6=idealaddone(nf,p5,x);
	p4=element_div(nf,(GEN)p6[1],p3moins);
/* p4 est l'inverse de p3moins modulo x */
	p3=element_mulmodideal(nf,p3plus,p4,x);
      }
      else p3=unnf;
      generator=(GEN)structarch[2];
      p5=gadd(gadd(zsigne(nf,p3,arch,nba),zsigne(nf,p3plus,arch,nba)),zsigne(nf,p3moins,arch,nba));
      p2=lift(gmul((GEN)structarch[3],p5));
      for(i=1;i<=nba;i++)
	if(signe((GEN)p2[i])) p3=element_mul(nf,p3,(GEN)generator[i]);
/* on a corrige pour que p3 ait la meme signature que p3plus/p3moins */
      basecl[j]=(long)p3;
    }
    p1=cgetg(4,17);p1[1]=(long)clh;
    mot=cgetg(c+1,17);for(i=1;i<=c;i++) mot[i]=coeff(met,i,i);
    p1[2]=(long)mot;p1[3]=(long)basecl;
  }
  else
  {
    p1=cgetg(3,17);p1[1]=(long)clh;
    mot=cgetg(c+1,17);for(i=1;i<=c;i++) mot[i]=coeff(met,i,i);
    p1[2]=(long)mot;
  }
  y[2]=(long)p1;
  tetpil=avma;return gerepile(av,tetpil,gcopy(y));
}

GEN
zidealstarinitjoinarch(GEN nf, GEN bid1, GEN arch, long nba)
{
  return zidealstarinitjoinarchall(nf,bid1,arch,nba,0);
}

GEN
zidealstarinitjoinarchgen(GEN nf, GEN bid1, GEN arch, long nba)
{
  return zidealstarinitjoinarchall(nf,bid1,arch,nba,1);
}

GEN
ideallist(GEN nf,long bound)
{
  byteptr ptdif=diffptr;
  long diff,prime=0,degk,lim,av0,av,tetpil,i,j,k,p2s,lfa,lp1;
  ulong q;
  GEN p,z,z2,pol,p1,p2,fa,pr,normp,ideal;

  nf=checknf(nf);if(bound<=0) return cgetg(1,17);
  av0=avma;p=cgeti(3);p[1]=evalsigne(1)+evallgef(3);
  z=cgetg(bound+1,17);for(i=1;i<=bound;i++) z[i]=lgetg(1,17);
  z2=cgetg(bound+1,17);
  pol=(GEN)nf[1];degk=lgef(pol)-3;
  p1=cgetg(2,17);z[1]=(long)p1;p1[1]=(long)idmat(degk);
  diff= *ptdif++;prime+=diff;if(!diff) err(primer1);
  av=avma;lim=(bot+avma)>>1;
  do
  {
    p[2]=prime;
    if(DEBUGLEVEL>=2) {fprintferr("%ld ",prime);flusherr();}
    fa=primedec(nf,p);lfa=lg(fa)-1;
    for(j=1;j<=lfa;j++)
    {
      pr=(GEN)fa[j];normp=gpui(p,(GEN)pr[4],0);
      if(gcmpgs(normp,bound)<=0)
      {
	z2=gcopy(z);
	q=p2s=itos(normp);ideal=pr;
	while((ulong)q<=(ulong)bound)
	{
	  for(i=q;i<=bound;i+=q)
	  {
	    p1=(GEN)z[i/q];lp1=lg(p1);
	    if(lp1>1)
	    {
	      p2=cgetg(lp1,17);
	      for(k=1;k<lp1;k++) p2[k]=(long)idealmul(nf,(GEN)p1[k],ideal);
	      z2[i]=(lg((GEN)z2[i])>1)?(long)concat((GEN)z2[i],p2):(long)p2;
	    }
	  }
	  q*=p2s;if((ulong)q<=(ulong)bound) ideal=idealmul(nf,ideal,pr);
	}
	z=z2;
      }
    }
    diff= *ptdif++;prime+=diff;if(!diff) err(primer1);
    if(avma<lim)
    {
      tetpil=avma;z=gerepile(av,tetpil,gcopy(z));
    }
  }
  while(prime<=bound);
  tetpil=avma;return gerepile(av0,tetpil,gcopy(z));
}

GEN
logunitmatrix(GEN nf,GEN funits,GEN racunit,GEN bid)
/* calcule la matrice des zinternallog des unites */
{
  long R,j,sizeh;
  GEN m,fa2,fa,arch;

  R=lg(funits)-1;
  m=cgetg(R+2,19);
  fa2=(GEN)bid[4];sizeh=lg((GEN)bid[5])-1;arch=(GEN)((GEN)bid[1])[2];
  fa=(GEN)bid[3];
  m[1]=(long)zinternallog(nf,fa2,sizeh,arch,fa,racunit,0);
  for(j=2;j<=R+1;j++) m[j]=(long)zinternallog(nf,fa2,sizeh,arch,fa,(GEN)funits[j-1],0);
  return m;
}

GEN
logunitmatrixarch(GEN nf,GEN funits,GEN racunit,GEN bid,long nba)
/* calcule la matrice des zinternallog des unites */
{
  long R,j;
  GEN m,liste,structarch,arch;

  R=lg(funits)-1;m=cgetg(R+2,19);arch=(GEN)((GEN)bid[1])[2];
  liste=(GEN)bid[4];structarch=(GEN)liste[lg(liste)-1];
  m[1]=(long)zsigne(nf,racunit,arch,nba);
  for(j=2;j<=R+1;j++) m[j]=(long)zsigne(nf,(GEN)funits[j-1],arch,nba);
  return lift(gmul((GEN)structarch[3],m));
}

GEN
vconcat(GEN Q1, GEN Q2)
/* concatenation verticale de Q1 et Q2. Ne cree pas le resultat. */
{
  long lc,lr,lx1,lx2,i,j;
  GEN m,p1,p2,p3;

  lc=lg(Q1);if(lc==1) return Q1;
  lx1=lg((GEN)Q1[1]);lx2=lg((GEN)Q2[1]);lr=lx1+lx2-1;
  m=cgetg(lc,19);
  for(j=1;j<lc;j++)
  {
    p1=cgetg(lr,18);m[j]=(long)p1;p2=(GEN)Q1[j];p3=(GEN)Q2[j];
    for(i=1;i<lx1;i++) p1[i]=p2[i];
    for(;i<lr;i++) p1[i]=p3[i-lx1+1];
  }
  return m;
}
  
GEN
ideallistzstarall(GEN bnf,long bound,long flun)
/* flun pair: pas de generateurs, impair: generateurs.
   flun>=2 unites, sinon pas. */
{
  byteptr ptdif=diffptr;
  long diff,prime=0,degk,lim,av0,av,tetpil,i,j,k,p2s,lfa,lp1,dec,flu;
  ulong q;
  GEN nf,p,z,z2,pol,p1,p2,p1u,p2u,fa,pr,normp,ideal,bidp,lu,lu2;
  GEN bigres,funits,racunit,embunit,y,sous,souslu,bid;

  flu=(flun>=2);flun=flun&1;
  if(flu)
  {
    checkbnf(bnf);
    nf=(GEN)bnf[7];bigres=(GEN)bnf[8];
    if(lg(bigres)!=5)
    {
      if(lg(bigres)!=7) err(talker,"incorrect big number field");
      else funits=(GEN)bigres[5];
    }
    else funits=(GEN)buchfu(bnf)[1];
    racunit=(GEN)((GEN)bigres[4])[2];
  }
  else nf=checknf(bnf);
  av0=avma;p=cgeti(3);p[1]=evalsigne(1)+evallgef(3);
  z=cgetg(bound+1,17);for(i=1;i<=bound;i++) z[i]=lgetg(1,17);
  z2=cgetg(bound+1,17);
  if(flu)
  {
    lu=cgetg(bound+1,17);for(i=1;i<=bound;i++) lu[i]=lgetg(1,17);
    lu2=cgetg(bound+1,17);
  }
  pol=(GEN)nf[1];degk=lgef(pol)-3;
  p1=cgetg(2,17);z[1]=(long)p1;p1[1]=(long)zidealstarinitall(nf,idmat(degk),flun);
  if(flu)
  {
    p2=cgetg(2,17);lu[1]=(long)p2;p2[1]=(long)logunitmatrix(nf,funits,racunit,(GEN)p1[1]);
  }
  diff= *ptdif++;prime+=diff;if(!diff) err(primer1);
  av=avma;lim=(bot+avma)>>1;
  do
  {
    p[2]=prime;
    if(DEBUGLEVEL>=2) {fprintferr("%ld ",prime);flusherr();}
    fa=primedec(nf,p);lfa=lg(fa)-1;
    for(j=1;j<=lfa;j++)
    {
      pr=(GEN)fa[j];normp=gpui(p,(GEN)pr[4],0);
      if(gcmpgs(normp,bound)<=0)
      {
	z2=gcopy(z);if(flu) lu2=gcopy(lu);
	q=p2s=itos(normp);ideal=pr;
	while((ulong)q<=(ulong)bound)
	{
	  bidp=zidealstarinitall(nf,ideal,flun);
	  if(flu) embunit=logunitmatrix(nf,funits,racunit,bidp);
	  for(i=q;i<=bound;i+=q)
	  {
	    p1=(GEN)z[i/q];lp1=lg(p1);if(flu) p1u=(GEN)lu[i/q];
	    if(lp1>1)
	    {
	      p2=cgetg(lp1,17);if(flu) p2u=cgetg(lp1,17);
	      for(k=1;k<lp1;k++)
	      {
		p2[k]=(long)zidealstarinitjoinall(nf,(GEN)p1[k],bidp,flun);
		if(flu) p2u[k]=(long)vconcat((GEN)p1u[k],embunit);
	      }
	      z2[i]=(lg((GEN)z2[i])>1)?(long)concat((GEN)z2[i],p2):(long)p2;
	      if(flu) lu2[i]=(lg((GEN)lu2[i])>1)?(long)concat((GEN)lu2[i],p2u):(long)p2u;
	    }
	  }
	  q*=p2s;
	  if((ulong)q<=(ulong)bound) ideal=idealmul(nf,ideal,pr);
	}
	z=z2;if(flu) lu=lu2;
      }
    }
    diff= *ptdif++;prime+=diff;if(!diff) err(primer1);
    if(avma<lim)
    {
      tetpil=avma;
      if(flu)
      {
	z=gcopy(z);lu=gcopy(lu);
	dec=lpile(av,tetpil,0)>>TWOPOTBYTES_IN_LONG;
	z+=dec;lu+=dec;
      }
      else z=gerepile(av,tetpil,gcopy(z));
    }
  }
  while(prime<=bound);
  if(flu)
  {
    for(i=1;i<lg(z);i++)
    {
      sous=(GEN)z[i];souslu=(GEN)lu[i];
      for(j=1;j<lg(sous);j++)
      {
	bid=(GEN)sous[j];souslu[j]=lmul((GEN)bid[5],(GEN)souslu[j]);
      }
    }
    tetpil=avma;y=cgetg(3,17);y[1]=lcopy(z);
    y[2]=lcopy(lu);
    return gerepile(av0,tetpil,y);
  }
  else {tetpil=avma;return gerepile(av0,tetpil,gcopy(z));}
}

GEN
ideallistzstar(GEN nf,long bound)
{
  return ideallistzstarall(nf,bound,0);
}

GEN
ideallistzstargen(GEN nf,long bound)
{
  return ideallistzstarall(nf,bound,1);
}

GEN
ideallistunit(GEN nf,long bound)
{
  return ideallistzstarall(nf,bound,2);
}

GEN
ideallistunitgen(GEN nf,long bound)
{
  return ideallistzstarall(nf,bound,3);
}

GEN
ideallistarchall(GEN nf,GEN list,GEN arch,long flun)
{
  long nba,i,j,lx,ly;
  GEN p1,z,sous;

  nba=0;for(i=1;i<lg(arch);i++) if(signe((GEN)arch[i])) nba++;
  lx=lg(list);z=cgetg(lx,17);
  for(i=1;i<lx;i++)
  {
    sous=(GEN)list[i];ly=lg(sous);p1=cgetg(ly,17);z[i]=(long)p1;
    for(j=1;j<ly;j++)
    {
      p1[j]=(long)zidealstarinitjoinarchall(nf,(GEN)sous[j],arch,nba,flun);
    }
  }
  return z;
}

GEN
ideallistarch(GEN nf, GEN list, GEN arch)
{
  return ideallistarchall(nf,list,arch,0);
}

GEN
ideallistarchgen(GEN nf, GEN list, GEN arch)
{
  return ideallistarchall(nf,list,arch,1);
}

GEN
ideallistunitarchall(GEN bnf,GEN list,GEN arch,long flun)
{
  long av=avma,tetpil,nba,i,j,lx,ly;
  GEN p1,p2,p3,racunit,funits,z,zold,lu,luold,nf,bigres,sous,souslu,embunit;
  GEN listnew;

  nba=0;for(i=1;i<lg(arch);i++) if(signe((GEN)arch[i])) nba++;
  checkbnf(bnf);nf=(GEN)bnf[7];bigres=(GEN)bnf[8];
  if(lg(bigres)!=5)
  {
    if(lg(bigres)!=7) err(talker,"incorrect big number field");
    else funits=(GEN)bigres[5];
  }
  else funits=(GEN)buchfu(bnf)[1];
  racunit=(GEN)((GEN)bigres[4])[2];
  zold=(GEN)list[1];luold=(GEN)list[2];
  lx=lg(zold);z=cgetg(lx,17);lu=cgetg(lx,17);
  for(i=1;i<lx;i++)
  {
    sous=(GEN)zold[i];souslu=(GEN)luold[i];ly=lg(sous);
    p1=cgetg(ly,17);z[i]=(long)p1;p2=cgetg(ly,17);lu[i]=(long)p2;
    for(j=1;j<ly;j++)
    {
      p3=zidealstarinitjoinarchall(nf,(GEN)sous[j],arch,nba,flun);
      p1[j]=(long)p3;embunit=logunitmatrixarch(nf,funits,racunit,p3,nba);
      p2[j]=lmul((GEN)p3[5],vconcat((GEN)souslu[j],embunit));
    }
  }
  tetpil=avma;
  listnew=cgetg(3,17);listnew[1]=lcopy(z);listnew[2]=lcopy(lu);
  return gerepile(av,tetpil,listnew);
}

GEN
ideallistunitarch(GEN bnf,GEN list,GEN arch)
{
  return ideallistunitarchall(bnf,list,arch,0);
}

GEN
ideallistunitarchgen(GEN bnf,GEN list,GEN arch)
{
  return ideallistunitarchall(bnf,list,arch,1);
}
