/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*                                                                 */
/*                     PLOT EN HAUTE RESOLUTION                    */
/*                                                                 */
/*                       copyright Babe Cool                       */
/*                                                                 */
/*                                                                 */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

# include "genpari.h"
#include <unistd.h>  /* we need fork() */

#ifdef HPPA
#ifndef __GNUC__
typedef char *caddr_t;
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#ifdef SOLARIS
pid_t fork(void);
#else
int fork();
#endif
#ifdef __cplusplus
}
#endif

Colormap PARI_Colormap;	/* colormap we are using */
XColor  *PARI_Colors;
XColor  *PARI_ExactColors;

static char *PARI_DefaultColors[8] = {
  " ",
  "sienna", /* Courbe impaires dans plothmult */
  "cornsilk",   
  "red",    /* Courbes, ou courbes paires dans plothmult */
  "black",                    
  "grey",                     
  "blue",   /* axes */
  "gainsboro",              
};

void
PARI_ColorSetUp(Display *display, char **Colors, int n)
{
  static int NotInitialised = 1;
  if (NotInitialised)
  {
    int i;
    NotInitialised = 0;
    PARI_Colormap = DefaultColormap(display, 0);
    PARI_Colors = (XColor *) malloc((n+1) * sizeof(XColor));
    PARI_ExactColors = (XColor *) malloc((n+1) * sizeof(XColor));
    for (i = 1; i < n; i++)
    {
      XAllocNamedColor(display, PARI_Colormap, Colors[i],
		       &PARI_ExactColors[i], &PARI_Colors[i]);
    }
  }
}

#define IDEC 90
#define JDEC 5

#define PLOTH_NUMPOINTS 1000
#define PLOTH2_NUMPOINTS 1500
#define PLOTHMULT_NUMPOINTS 2000

#define PLOTHR_NUMPOINTS 10
#define PLOTHR_MAXDEPTH 10
#define PLOTHR_PRECISION 0.001

#define PLOTH2R_NUMPOINTS 10
#define PLOTH2R_MAXDEPTH 15
#define PLOTH2R_PRECISION 0.001

#ifdef SONY
typedef int (*XErrorHandler) (
#if NeedFunctionPrototypes
  Display*	
  XErrorEvent*
#endif
);

extern XErrorHandler XSetErrorHandler (
#if NeedFunctionPrototypes
  XErrorHandler	
#endif
);

typedef int (*XIOErrorHandler) ( 
#if NeedFunctionPrototypes
  Display*	
#endif
);

extern XIOErrorHandler XSetIOErrorHandler (
#if NeedFunctionPrototypes
  XIOErrorHandler
#endif
);
#endif

int
xerror()
{
  err(ploter1);return 0;
}

/* List of points, each containing two double coordinates */
struct dblPointListElement
{
  struct dblPointListElement *next;
  
  double x;
  double y;
};

struct dblPointList
{
  struct dblPointListElement *FirstElement,*LastElement;
  long NumberOfElements;
  double xsml,xbig,ysml,ybig;
};

/* Initialise the list */
static void InitdblPointList(struct dblPointList *l)
{
  l->FirstElement=0;
  l->LastElement=0;
  l->NumberOfElements=0;
  l->xsml = l->xbig = l->ysml = l->ybig=0.0;
}

/* append a point to a list */
static void dblPointListAppend
  (struct dblPointList *first,struct dblPointList *l,double x,double y)
{
  struct dblPointListElement *NewElement
    =(struct dblPointListElement*)malloc(sizeof(struct dblPointListElement));

  if (NewElement==0)
    /* out of memory! */
    return;
  
  if (l->LastElement==0)
    l->FirstElement=l->LastElement=NewElement;
  else
    l->LastElement=l->LastElement->next=NewElement;
  l->LastElement->next=0;
  l->LastElement->x=x;
  l->LastElement->y=y;
  (l->NumberOfElements)++;

  if (x < first->xsml) first->xsml=x;
  if (x > first->xbig) first->xbig=x;
  if (y < first->ysml) first->ysml=y;
  if (y > first->ybig) first->ybig=y;
}

/* correct the boundaries, that means assure, xsml<xbig and ysml<ybig */
static void correctBoundaries(struct dblPointList *l)
{
  if (l->ybig<=l->ysml)
    {
      l->ybig=l->ybig+1.0;
      l->ysml=l->ysml-1.0;
    }
  if (l->xbig<=l->xsml)
    {
      l->xbig=l->xbig+1.0;
      l->xsml=l->xsml-1.0;
    }
}

/* compute xpoints from double point list */
static void ComputeXPointsFromdlbPointList
(struct dblPointList *first,struct dblPointList *l,XPoint *xpoints,
long width,long height,long xoffset,long yoffset)
{
  int i;
  struct dblPointListElement *e;
  
  for(i = 0, e=l->FirstElement; 
      i < l->NumberOfElements && e!=0; 
      i++,e=e->next)
  {
    xpoints[i].x = 
      xoffset+(long)((e->x - first->xsml)/(first->xbig - first->xsml)*width);
    xpoints[i].y = 
      yoffset+(long)((e->y - first->ybig)/(first->ysml - first->ybig)*height);
  }
}

/* Delete all listelements */
static void DeleteAllPoints(struct dblPointList *l)
{
  struct dblPointListElement *e1,*e2;
  for(e1=l->FirstElement; e1!=0; e1=e2)
  {
    e2=e1->next;
    free(e1);
  }
}
      
void spawn_viewer(struct dblPointList **l, long nc, char *title, char line)
{
  int fork_result,i;

  fork_result=fork();
  if (fork_result==-1) /* fork failed */
  {
    fprintf(stderr,"sorry, cannot fork\n");
    return;
  }
  if (fork_result>0) return; /* parent process */
  
  /* child process */
  {
    XPoint *xpoints;
    Display *display;
    Window win;
    XSizeHints size_hints;
    XEvent report;
    GC gc;
    XFontStruct *font_info;
    XSetWindowAttributes attrib;
    int screen;
    unsigned int width,height,color;
    double iz0,jz0;
    long iz,jz,is,js;

    char c1[20],c2[20],c3[20],c4[20];
    
    correctBoundaries(l[0]);

    /* setup the x parameters */
    if (!(display = XOpenDisplay(NULL)))
    {
      err(ploter2);
      exit(1);
    }
    if (!(font_info = XLoadQueryFont(display, "9x15")))
    {
      err(ploter3);
      exit(1);
    }
    xpoints=(XPoint*) malloc(l[0]->NumberOfElements*sizeof(XPoint));
    if (!xpoints)
    {
      err(talker,"Not enough memory in spawn_viewer");
      exit(1);
    }
    XSetErrorHandler((XErrorHandler)xerror);
    XSetIOErrorHandler((XIOErrorHandler)xerror);
    PARI_ColorSetUp(display,PARI_DefaultColors, 8);
    
    screen = DefaultScreen(display);
    width = DisplayWidth(display, screen) - 40;
    height = DisplayHeight(display, screen) - 60;

    win = XCreateSimpleWindow
      (display, RootWindow(display, screen), 0, 0, width, height, 
       4, BlackPixel(display, screen), WhitePixel(display, screen));

    size_hints.flags = PPosition | PSize;
    size_hints.x = 0; size_hints.y = 0;
    size_hints.width = width; size_hints.height = height;

    XSetStandardProperties
     (display, win, title, NULL, None, NULL, 0, &size_hints);

    XSelectInput
     (display,win,ExposureMask|ButtonPressMask|StructureNotifyMask);

    /* enable backing-store */
    attrib.backing_store = Always;
    attrib.backing_planes = AllPlanes;
    XChangeWindowAttributes(display,win,CWBackingStore|CWBackingPlanes,&attrib);

    gc = XCreateGC(display, win, 0, NULL);
    XSetFont(display, gc, font_info->fid);

    XMapWindow(display, win);
    
    /* compute the xpoints from the point list */
    is = width - IDEC-5; js = height - JDEC-20;
    iz0=(- l[0]->xsml)/(l[0]->xbig - l[0]->xsml);
    jz0=(- l[0]->ybig)/(l[0]->ysml - l[0]->ybig);
    iz = (long) (iz0*is); jz = (long) (jz0*js);
    
    /* setup the labels */
    sprintf(c1,"%9.3f",l[0]->ybig);
    sprintf(c2,"%9.3f",l[0]->ysml);
    sprintf(c3,"%9.3f",l[0]->xsml);
    sprintf(c4,"%9.3f",l[0]->xbig);

    while(1)
    {
      Window WinRoot;
      int qxpos,qypos;
      unsigned int bwidth,dep;

      XNextEvent(display, &report);
      switch(report.type)
      {
	case ButtonPress:
	{
	  XUnloadFont(display, font_info->fid);
	  XFreeGC(display, gc);
	  XCloseDisplay(display);
	  for(i=0;i<nc;i++)
	  {
	    DeleteAllPoints(l[i]);
	    free(l[i]);
	  }
	  exit(0);
	}

	case ConfigureNotify:
	{
	  /* get window size */
	  XGetGeometry (
	    display,win, &WinRoot, &qxpos,&qypos, &width,&height, &bwidth,&dep
	  );

	  /* recompute scale */
	  is = width - IDEC-5; js = height - JDEC-20;
	  iz = (long) (iz0*is); jz = (long) (jz0*js);
	}

	case Expose:
	{
	  XSetForeground(display, gc, BlackPixel(display, screen));
	  XDrawRectangle(display, win, gc, IDEC, JDEC, is, js);
	  XDrawString(display, win, gc, 0, 15, c1, strlen(c1));
	  XDrawString(display, win, gc, 0, height-30, c2, strlen(c2));
	  XDrawString(display, win, gc, IDEC-20, height-5, c3, strlen(c3));
	  XDrawString(display, win, gc, width-90, height-5, c4, strlen(c4));
	  XSetForeground(display, gc, PARI_Colors[6].pixel);
	  if (jz<=js)
	    XDrawLine(display, win, gc, IDEC, JDEC+jz, IDEC+is, JDEC+jz);
	  if (iz>=0)
	    XDrawLine(display, win, gc, IDEC+iz, JDEC, IDEC+iz, js+JDEC);

	  for (i=0;i<nc;i++)
	  {
	    xpoints=(XPoint*)
	      realloc(xpoints,l[i]->NumberOfElements*sizeof(XPoint));
	    if (!xpoints)
	    {
	      err(talker,"Not enough memory in spawn_viewer");
	      exit(1);
	    }
	    color=((i&1)?1:3);
	    XSetForeground(display, gc, PARI_Colors[color].pixel);
	    ComputeXPointsFromdlbPointList(l[0],l[i],xpoints,is,js,IDEC,JDEC);
	    if (line)
	      XDrawLines(display, win, gc, xpoints, l[i]->NumberOfElements,0);
	    else
	      XDrawPoints(display, win, gc, xpoints, l[i]->NumberOfElements,0);
	  }
	}
      }
    }
  }
}


GEN
ploth(entree *ep, GEN a, GEN b, char *ch, long prec)
{
  long av = avma,av2,i,sig;
  GEN x,dx,y;
  struct dblPointList *pl;

  sig=gcmp(b,a); if(!sig) return gnil;
  if(sig<0) {x=a;a=b;b=x;}
  
  pl=(struct dblPointList*)malloc(sizeof(struct dblPointList));
  if (!pl) err(talker,"Not enough memory in ploth");
  InitdblPointList(pl);
  pl->xsml=gtodouble(a); pl->xbig=gtodouble(b);

  y=cgetr(3);
  newvalue(ep,cgetr(prec)); 
  x=(GEN)(ep->value);
  gaffect(a,x);
  
  dx=gdivgs(gsub(b,a),PLOTH_NUMPOINTS-1);
  av2=avma;
  for(i=0;i<PLOTH_NUMPOINTS;i++)
  {
    gaffect(lisexpr(ch), y);
    dblPointListAppend(pl,pl,rtodbl(x),rtodbl(y));

    gaddz(x,dx,x);
    avma=av2;
  }

  spawn_viewer(&pl,1,"ploth",1);
  
  avma = av;
  killvalue(ep);
  return gnil;
}

static void plothr_recursion
(struct dblPointList *pl,char *ch,GEN x,
GEN xleft,GEN yleft, 
GEN xright,GEN yright, 
int depth)
{
  GEN xx,yy;
  long av=avma;
  
  if (depth==PLOTHR_MAXDEPTH) return;

  yy=cgetr(3); xx=gdivgs(gadd(xleft,xright),2);
  gaffect(xx,x); gaffect(lisexpr(ch), yy);

  if (fabs(rtodbl(yleft)+rtodbl(yright)-2*rtodbl(yy)) < PLOTHR_PRECISION)
    return;

  plothr_recursion(pl,ch,x, xleft,yleft, xx,yy, depth+1);
  dblPointListAppend(pl,pl,rtodbl(xx),rtodbl(yy));
  plothr_recursion(pl,ch,x, xx,yy, xright,yright, depth+1);

  avma=av;
}

GEN
plothr(entree *ep, GEN a, GEN b, char *ch, long prec)
{
  long av = avma,av2,i,sig;
  GEN x,dx,xleft,yleft,xright,yright;
  struct dblPointList *pl;
  
  sig=gcmp(b,a); if(!sig) return gnil;
  if(sig<0){x=a;a=b;b=x;}

  pl=(struct dblPointList*) malloc(sizeof(struct dblPointList));
  if (!pl) err(talker,"Not enough memory in plothr");
  InitdblPointList(pl);
  pl->xsml=gtodouble(a); pl->xbig=gtodouble(b);

  newvalue(ep,cgetr(prec)); x=(GEN)(ep->value);
  gaffect(a,x);

  xleft=cgetr(3); xright=cgetr(3);
  yleft=cgetr(3); yright=cgetr(3);
  gaffect(a,xleft);
  dx=gdivgs(gsub(b,a),PLOTHR_NUMPOINTS);

  av2=avma;
  gaffect(a,xleft);gaffect(xleft,x);gaffect(lisexpr(ch),yleft);
  for(i=0; i<PLOTHR_NUMPOINTS; i++)
  {
    gaddz(xleft,dx,xright);
    gaffect(xright,x); gaffect(lisexpr(ch),yright);

    dblPointListAppend(pl,pl,rtodbl(xleft),rtodbl(yleft));

    plothr_recursion(pl,ch,x,xleft,yleft,xright,yright,0);
    avma=av2;
    gaffect(xright,xleft);gaffect(yright,yleft);
  }
  dblPointListAppend(pl,pl,rtodbl(xright),rtodbl(yright));

  spawn_viewer(&pl,1,"plothr",1);
  
  avma = av;
  killvalue(ep);
  return gnil;
}

GEN
ploth2(entree *ep, GEN a, GEN b, char *ch, long prec)
{
  long av = avma,av2,i,sig;
  GEN p1,t,dt,x,y;
  struct dblPointList *pl;
  

  sig=gcmp(b,a); if(!sig) return gnil;
  if(sig<0) {x=a;a=b;b=x;}

  pl=(struct dblPointList*)malloc(sizeof(struct dblPointList));
  if (!pl) err(talker,"Not enough memory in ploth2");
  InitdblPointList(pl);
  
  x=cgetr(3);
  y=cgetr(3);
  newvalue(ep,cgetr(prec)); 
  t=(GEN)(ep->value);
  gaffect(a,t);
  
  dt=gdivgs(gsub(b,a),PLOTH2_NUMPOINTS-1);
  av2=avma;
  for(i=0;i<PLOTH2_NUMPOINTS;i++)
  {
    p1 = lisexpr(ch);
    gaffect((GEN)p1[1], x); 
    gaffect((GEN)p1[2], y);

    dblPointListAppend(pl,pl,rtodbl(x),rtodbl(y));

    gaddz(t,dt,t);
    avma=av2;
  }

  spawn_viewer(&pl,1,"ploth2",1);
  
  avma = av;
  killvalue(ep);
  return gnil;
}

static void ploth2r_recursion
(struct dblPointList *pl,char *ch,GEN t, 
GEN tleft,GEN xleft,GEN yleft, 
GEN tright,GEN xright,GEN yright, 
int depth)
{
  GEN tt,xx,yy, p1;
  long av;
  
  av=avma;
  
  if (depth==PLOTH2R_MAXDEPTH) return;

  xx=cgetr(3);
  yy=cgetr(3);
  
  tt=gdivgs(gadd(tleft,tright),2);
  gaffect(tt,t);
  
  p1=lisexpr(ch);
  gaffect((GEN)p1[1],xx);
  gaffect((GEN)p1[2],yy);

  if ( 
      fabs(rtodbl(xleft)+rtodbl(xright)-2*rtodbl(xx)) < PLOTH2R_PRECISION &&
      fabs(rtodbl(yleft)+rtodbl(yright)-2*rtodbl(yy)) < PLOTH2R_PRECISION
      )
    return;

  ploth2r_recursion(pl,ch,t, tleft,xleft,yleft, tt,xx,yy, depth+1);
  dblPointListAppend(pl,pl,rtodbl(xx),rtodbl(yy));
  ploth2r_recursion(pl,ch,t, tt,xx,yy, tright,xright,yright, depth+1);

  avma=av;
}

GEN
ploth2r(entree *ep, GEN a, GEN b, char *ch, long prec)
{
  long av = avma,av2,i,sig;
  GEN p1,t,dt, tleft,xleft,yleft, tright,xright,yright;
  struct dblPointList *pl; 
  sig=gcmp(b,a); if(!sig) return gnil;
  if(sig<0) {t=a;a=b;b=t;}

  pl=(struct dblPointList*)malloc(sizeof(struct dblPointList));
  if (!pl) err(talker,"Not enough memory in ploth2r");
  InitdblPointList(pl);

  newvalue(ep,cgetr(prec)); t=(GEN)(ep->value);
  gaffect(a,t);
  av2=avma;

  xleft=cgetr(3);
  xright=cgetr(3);
  yleft=cgetr(3);
  yright=cgetr(3);
  tleft=cgetr(prec);
  tright=cgetr(prec);
  gaffect(a,tleft);
  dt=gdivgs(gsub(b,a),PLOTH2R_NUMPOINTS);

  av2=avma;
  for(i=0; i<PLOTH2R_NUMPOINTS; i++)
    {
      if (i==0)
	{
	  gaffect(a,tleft);
	  gaffect(tleft,t);
	  p1=lisexpr(ch);
	  gaffect((GEN)p1[1],xleft);
	  gaffect((GEN)p1[2],yleft);
	}
      else
	{
	  gaffect(tright,tleft);
	  gaffect(xright,xleft);
	  gaffect(yright,yleft);
	}
      
      gaddz(tleft,dt,tright);
      gaffect(tright,t);
      p1=lisexpr(ch);
      gaffect((GEN)p1[1],xright);
      gaffect((GEN)p1[2],yright);

      dblPointListAppend(pl,pl,rtodbl(xleft),rtodbl(yleft));

      ploth2r_recursion
	(pl,ch,t, tleft,xleft,yleft, tright,xright,yright, 0);
      avma=av2;
    }
  dblPointListAppend(pl,pl,rtodbl(xright),rtodbl(yright));

  spawn_viewer(&pl,1,"ploth2r",1);
  
  avma = av;
  killvalue(ep);
  return gnil;
}

GEN
plothraw(GEN listx, GEN listy)
{
  long av=avma,av2,i,lx;
  struct dblPointList *pl;

  if((typ(listx)<17)||(typ(listx)>18)||(typ(listy)<17)||(typ(listy)>18))
     err(ploter4);
  lx=lg(listx);
  if(lg(listy)!=lx) err(ploter5);
  if(lx==1) return gnil;

  pl=(struct dblPointList*)malloc(sizeof(struct dblPointList));
  if (!pl) err(ploter6);
  InitdblPointList(pl);

  pl->xsml=pl->xbig=gtodouble((GEN)listx[1]);
  pl->ysml=pl->ybig=gtodouble((GEN)listy[1]);

  av2=avma;
  for(i = 1; i < lx; i++)
  {
    dblPointListAppend(pl,pl,gtodouble((GEN)listx[i]),gtodouble((GEN)listy[i]));
    avma=av2;
  }

  spawn_viewer(&pl,1,"plothraw",0);

  avma=av;
  return gnil;
}

GEN
rectdraw(GEN list)
{
  long *e,*p1,*ptx,*pty,*numpoints,*numtexts,*xtexts,*ytexts;
  long n,i,j,x0,y0,av=avma;
  long a,b,nd[10],ne;
  char **texts;

  Display *display;
  Window win;
  XSizeHints size_hints;
  XEvent report;
  GC gc;
  XFontStruct *font_info;
  int screen, width, height;
  XPoint *points,**lines;
  XSegment *segments;
  XRectangle *rectangles;

  if(typ(list)!=17) err(rploter3);
  n=lg(list)-1;if(n%3) err(rploter4);
  n=n/3;if(!n) return gnil;
  if (!(display = XOpenDisplay(NULL))) err(ploter2);
  if (!(font_info = XLoadQueryFont(display, "9x15"))) err(ploter3);
  XSetErrorHandler((XErrorHandler)xerror);
  XSetIOErrorHandler((XIOErrorHandler)xerror);
  screen = DefaultScreen(display);
  width = DisplayWidth(display, screen) - 40;
  height = DisplayHeight(display, screen) - 60; 
  nd[0]=nd[1]=nd[2]=nd[3]=nd[4]=nd[5]=nd[6]=0;
  for(i=0;i<n;i++)
    {
      if(typ((GEN)list[3*i+1])!=1) err(rploter5);
      ne=itos((GEN)list[3*i+1]);if((ne<0)||(ne>15)) err(rploter2);
      e=rectgraph[ne];
      p1=(long*)e[0];while((long)p1) 
	{
	  if(p1[1]!=4) nd[p1[1]]++;
	  else nd[1]+=p1[2];
	  p1=(long*)p1[0];
	}
    }
  points=(XPoint*)malloc(nd[1]*sizeof(XPoint));
  segments=(XSegment*)malloc(nd[2]*sizeof(XSegment));
  rectangles=(XRectangle*)malloc(nd[3]*sizeof(XRectangle));
  lines=(XPoint**)malloc(nd[5]*sizeof(XPoint*));
  numpoints=(long*)malloc(nd[5]*sizeof(long));
  texts=(char**)malloc(nd[6]*sizeof(char*));
  numtexts=(long*)malloc(nd[6]*sizeof(long));
  xtexts=(long*)malloc(nd[6]*sizeof(long));
  ytexts=(long*)malloc(nd[6]*sizeof(long));
  nd[1]=nd[2]=nd[3]=nd[5]=nd[6]=0;
  for(i=0;i<n;i++)
    {
      e=rectgraph[itos((GEN)list[3*i+1])];x0=list[3*i+2];y0=list[3*i+3];
      if((typ((GEN)x0)!=1)||(typ((GEN)y0)!=1)) err(rploter5);
      x0=itos((GEN)x0);y0=itos((GEN)y0);
      p1=(long*)e[0];
      while((long)p1)
	{
	  switch(p1[1])
	    {
	    case 1: 
	      points[nd[1]].x=p1[2]+x0;
	      points[nd[1]].y=p1[3]+y0;
	      nd[1]++;break;
	    case 2:
	      segments[nd[2]].x1=p1[2]+x0;
	      segments[nd[2]].y1=p1[3]+y0;
	      segments[nd[2]].x2=p1[4]+x0;
	      segments[nd[2]].y2=p1[5]+y0;
	      nd[2]++;break;
	    case 3:
	      a=rectangles[nd[3]].x=p1[2]+x0;
	      b=rectangles[nd[3]].y=p1[3]+y0;
	      rectangles[nd[3]].width=p1[4]+x0-a;
	      rectangles[nd[3]].height=p1[5]+y0-b;
	      nd[3]++;break;
	    case 4:
	      ptx=(long*)p1[3];pty=(long*)p1[4];
	      for(j=0;j<p1[2];j++)
		{
		  points[nd[1]+j].x=ptx[j]+x0;
		  points[nd[1]+j].y=pty[j]+y0;
		}
	      nd[1]+=p1[2];break;
	    case 5:
	      ptx=(long*)p1[3];pty=(long*)p1[4];
	      numpoints[nd[5]]=p1[2];
	      lines[nd[5]]=(XPoint*)malloc(p1[2]*sizeof(XPoint));
	      for(j=0;j<p1[2];j++)
		{
		  lines[nd[5]][j].x=ptx[j]+x0;
		  lines[nd[5]][j].y=pty[j]+y0;
		}
	      nd[5]++;break;
	    case 6: 
	      texts[nd[6]]=(char*)p1[3];numtexts[nd[6]]=p1[2];
	      xtexts[nd[6]]=p1[4]+x0;ytexts[nd[6]]=p1[5]+y0;
	      nd[6]++;break;
	    default: break;
	    }
	  p1=(long*)p1[0];
	}
    }
  win= XCreateSimpleWindow(display, RootWindow(display, screen), 0, 0, width, height, 4, BlackPixel(display, screen), WhitePixel(display, screen));
  size_hints.flags = PPosition | PSize;
  size_hints.x = 0;
  size_hints.y = 0;
  size_hints.width = width;
  size_hints.height = height;
  
  XSetStandardProperties(display, win, "rectplot", NULL, None, NULL, 0, &size_hints);
  XSelectInput(display, win, ExposureMask | ButtonPressMask);
  gc = XCreateGC(display, win, 0, NULL);
  XSetFont(display, gc, font_info->fid);
  XSetForeground(display, gc, BlackPixel(display, screen)); 
  XMapWindow(display, win);
  for(;;)
    {
      XNextEvent(display, &report);
      if (report.type != Expose) break;
      while (XCheckTypedEvent(display, Expose, &report));
      if(nd[1]) XDrawPoints(display, win, gc, points, nd[1], 0);
      if(nd[2]) XDrawSegments(display, win, gc, segments, nd[2]);
      if(nd[3]) XDrawRectangles(display, win, gc, rectangles, nd[3]);
      for(i=0;i<nd[5];i++) XDrawLines(display,win,gc,lines[i],numpoints[i],0);
      for(i=0;i<nd[6];i++) XDrawString(display,win,gc,xtexts[i],ytexts[i],texts[i],numtexts[i]);
    }
  XUnloadFont(display, font_info->fid);
  XFreeGC(display, gc);
  free(points);free(segments);free(rectangles);
  free(numpoints);for(i=0;i<nd[5];i++) free(lines[i]);
  free(lines);free(texts);free(numtexts);free(xtexts);free(ytexts);
  XCloseDisplay(display);
  avma = av;return gnil;
}

GEN
plothmult(entree *ep, GEN a, GEN b, char *ch, long prec)
{
  long av = avma,av2,i,j,sig,nc;
  GEN p1,x,y,dx;
  struct dblPointList **pl;

  sig=gcmp(b,a); if(!sig) {abort();return gnil;}
  if(sig<0) {x=a;a=b;b=x;}
  
  newvalue(ep,cgetr(prec)); x=(GEN)(ep->value);
  gaffect(a,x);
  
  p1=lisexpr(ch);
  if(typ(p1)!=17) err(talker,"not a row vector in plothmult");
  nc=lg(p1);if(nc<=1) {avma=av;return gnil;}
  y=cgetg(nc,17); nc--;

  pl=(struct dblPointList**)malloc(nc*sizeof(struct dblPointList*));
  if (!pl) err(ploter6);
  for (i=0;i<nc;)
  {
    pl[i]=(struct dblPointList*)malloc(sizeof(struct dblPointList));
    if (!pl[i]) err(ploter6);
    InitdblPointList(pl[i]);
    i++; y[i]=lgetr(3);
  }
  pl[0]->xsml=gtodouble(a); pl[0]->xbig=gtodouble(b);

  dx=gdivgs(gsub(b,a),PLOTHMULT_NUMPOINTS-1);
  av2=avma;
  for(i=0;i<PLOTHMULT_NUMPOINTS;i++)
  {
    gaffect(lisexpr(ch),y);
    for(j=0;j<nc;j++)
    {
      dblPointListAppend(pl[0],pl[j],rtodbl(x),rtodbl((GEN)y[j+1]));
      avma=av2;
    }
    gaddz(x,dx,x);
  }
  spawn_viewer(pl,nc,"plothmult",1);
  avma = av;
  killvalue(ep);
  return gnil;
}
